#!/usr/bin/python3
"""Configuration file writer for sim_telarray."""

import logging
from copy import deepcopy
from pathlib import Path

import astropy.units as u
import numpy as np

import simtools.utils.general as gen
import simtools.version
from simtools.io import ascii_handler
from simtools.simtel.pulse_shapes import generate_pulse_from_rise_fall_times
from simtools.utils import names

logger = logging.getLogger(__name__)


def sim_telarray_random_seeds(seed, number):
    """
    Generate random seeds to be used in sim_telarray.

    Parameters
    ----------
    seed: int
        Seed for the random number generator.
    number: int
        Number of random seeds to generate.

    Returns
    -------
    list
        List of random seeds.
    """
    rng = np.random.default_rng(seed)
    max_int32 = np.iinfo(np.int32).max  # sim_telarray requires 32 bit integers
    return list(rng.integers(low=1, high=max_int32, size=number, dtype=np.int32))


class SimtelConfigWriter:
    """
    SimtelConfigWriter writes sim_telarray configuration files.

    It is designed to be used by model classes (TelescopeModel and ArrayModel) only.

    Parameters
    ----------
    site: str
        South or North.
    model_version: str
        Model version.
    telescope_model_name: str
        Telescope model name.
    layout_name: str
        Layout name.
    label: str
        Instance label. Important for output file naming.
    simtel_path: str or Path
        Path to the sim_telarray installation directory.
    """

    TAB = " " * 3

    def __init__(
        self,
        site,
        model_version,
        layout_name=None,
        telescope_model_name=None,
        telescope_design_model=None,
        label=None,
        simtel_path=None,
    ):
        """Initialize SimtelConfigWriter."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimtelConfigWriter")

        self._site = site
        self._model_version = model_version
        self._label = label
        self._layout_name = layout_name
        self._telescope_model_name = telescope_model_name
        self._telescope_design_model = telescope_design_model
        self._simtel_path = simtel_path

    def write_telescope_config_file(
        self, config_file_path, parameters, telescope_name=None, telescope_design_model=None
    ):
        """
        Write the sim_telarray config file for a single telescope.

        Parameters
        ----------
        config_file_path: str or Path
            Path of the file to write on.
        parameters: dict
            Model parameters
        telescope_name: str
            Name of the telescope (use self._telescope_model_name if None)
        telescope_design_model: str
            Telescope design model.
        """
        self._logger.debug(f"Writing telescope config file {config_file_path}")

        simtel_par = self._get_parameters_for_sim_telarray(parameters, config_file_path)
        telescope_name = telescope_name or self._telescope_model_name
        _telescope_design_model = telescope_design_model or self._telescope_design_model

        with open(config_file_path, "w", encoding="utf-8") as file:
            self._write_header(file, "TELESCOPE CONFIGURATION FILE")

            file.write("#ifdef TELESCOPE\n")
            file.write(f"   echo Configuration for {telescope_name} - TELESCOPE $(TELESCOPE)\n")
            file.write("#endif\n\n")

            for simtel_name, simtel_value in simtel_par.items():
                file.write(f"{simtel_name} = {self._get_value_string_for_simtel(simtel_value)}\n")
            for meta in self._get_sim_telarray_metadata(
                "telescope",
                parameters,
                telescope_name,
                telescope_design_model=_telescope_design_model,
            ):
                file.write(f"{meta}\n")

    @staticmethod
    def write_lightpulse_table_gauss_expconv(
        file_path,
        width_ns=None,
        exp_decay_ns=None,
        dt_ns=0.1,
        rise_range=(0.1, 0.9),
        fall_range=(0.9, 0.1),
        fadc_sum_bins=None,
        time_margin_ns=10.0,
    ):
        """Write a pulse table for a Gaussian convolved with a causal exponential.

        Parameters
        ----------
        file_path : str or pathlib.Path
            Destination path of the ASCII pulse table to write. Parent directory must exist.
        width_ns : float
            Target rise time in ns between the fractional levels defined by ``rise_range``.
            Defaults correspond to 10-90% rise time.
        exp_decay_ns : float
            Target fall time in ns between the fractional levels defined by ``fall_range``.
            Defaults correspond to 90-10% fall time.
        dt_ns : float, optional
            Time sampling step in ns for the generated pulse table. Default is 0.1.
        rise_range : tuple[float, float], optional
            Fractional amplitude bounds (low, high) for rise-time definition. Default (0.1, 0.9).
        fall_range : tuple[float, float], optional
            Fractional amplitude bounds (high, low) for fall-time definition. Default (0.9, 0.1).
        fadc_sum_bins : int
            Length of the FADC integration window (treated as ns here) used to derive
            the internal time sampling window of the solver as [-(margin), bins + margin].
        time_margin_ns : float, optional
            Margin in ns to add to both ends of the FADC window when ``fadc_sum_bins`` is given.
            Default is 5.0 ns.

        Returns
        -------
        pathlib.Path
            The path to the created pulse table file.

        Notes
        -----
        The underlying model is a Gaussian convolved with a causal exponential. The model
        parameters (sigma, tau) are solved such that the normalized pulse matches the requested
        rise and fall times. The pulse is normalized to a peak amplitude of 1.
        """
        if width_ns is None or exp_decay_ns is None:
            raise ValueError("width_ns (rise 10-90) and exp_decay_ns (fall 90-10) are required")
        logger.info(
            f"Generating lightpulse table with rise10-90={width_ns} ns, "
            f"fall90-10={exp_decay_ns} ns, dt={dt_ns} ns"
        )
        width = float(fadc_sum_bins)
        t_start_ns = -abs(time_margin_ns + width)
        t_stop_ns = +abs(time_margin_ns + width)
        t, y = generate_pulse_from_rise_fall_times(
            width_ns,
            exp_decay_ns,
            dt_ns=dt_ns,
            rise_range=rise_range,
            fall_range=fall_range,
            t_start_ns=t_start_ns,
            t_stop_ns=t_stop_ns,
            center_on_peak=True,
        )

        return SimtelConfigWriter._write_ascii_pulse_table(file_path, t, y)

    @staticmethod
    def _write_ascii_pulse_table(file_path, t, y):
        """Write two-column ASCII pulse table."""
        with open(file_path, "w", encoding="utf-8") as fh:
            fh.write("# time[ns] amplitude\n")
            for ti, yi in zip(t, y):
                fh.write(f"{ti:.6f} {yi:.8f}\n")
        return Path(file_path)

    def _get_parameters_for_sim_telarray(self, parameters, config_file_path):
        """
        Convert parameter dictionary to sim_telarray configuration file format.

        Accounts for differences between the data models for sim_telarray configuration
        and the simulation models.

        Parameters
        ----------
        parameters: dict
            Model parameters.
        config_file_path: str or Path
            Path of the file to write on.

        Returns
        -------
        dict
            Model parameters in sim_telarray format.
        """
        simtel_par = {}
        for par, value in parameters.items():
            simtel_name, simtel_value = self._convert_model_parameters_to_simtel_format(
                names.get_simulation_software_name_from_parameter_name(
                    par, software_name="sim_telarray"
                ),
                value["value"],
                config_file_path,
                None,
            )
            if simtel_name:
                simtel_par[simtel_name] = simtel_value
        if "stars" not in parameters:  # sim_telarray requires 'stars' to be set
            simtel_par["stars"] = None

        return self._get_flasher_parameters_for_sim_telarray(parameters, simtel_par)

    def _get_flasher_parameters_for_sim_telarray(self, parameters, simtel_par):
        """
        Combine flasher pulse time parameters into a single parameter.

        Takes into account that sim_telarray expects a single parameter with a specific name.

        Parameters
        ----------
        parameters: dict
            Model parameters.
        simtel_par: dict
            Model parameters in sim_telarray format.

        Returns
        -------
        dict
            Model parameters in sim_telarray format including flasher parameters.

        """
        if "flasher_pulse_shape" not in parameters and "flasher_pulse_width" not in parameters:
            return simtel_par

        mapping = {
            "gauss": "laser_pulse_sigtime",
            "tophat": "laser_pulse_twidth",
        }

        shape = parameters.get("flasher_pulse_shape", {}).get("value", "").lower()
        if "exponential" in shape:
            simtel_par["laser_pulse_exptime"] = parameters.get("flasher_pulse_exp_decay", {}).get(
                "value", 0.0
            )
        else:
            simtel_par["laser_pulse_exptime"] = 0.0

        width = parameters.get("flasher_pulse_width", {}).get("value", 0.0)

        simtel_par.update(dict.fromkeys(mapping.values(), 0.0))
        if shape == "gauss-exponential":
            simtel_par["laser_pulse_sigtime"] = width
        elif shape in mapping:
            simtel_par[mapping[shape]] = width
        else:
            self._logger.warning(f"Flasher pulse shape '{shape}' without width definition")

        return simtel_par

    def _get_value_string_for_simtel(self, value):
        """
        Return a value string for simtel.

        Parameters
        ----------
        value: any
            Value to convert to string.

        Returns
        -------
        str
            Value string for simtel.
        """
        value = "none" if value is None else value  # simtel requires 'none'
        if isinstance(value, bool):
            value = 1 if value else 0
        elif isinstance(value, list | np.ndarray):
            value = gen.convert_list_to_string(value, shorten_list=True)
        return value

    def _get_sim_telarray_metadata(
        self,
        config_type,
        model_parameters,
        telescope_model_name,
        additional_metadata=None,
        telescope_design_model=None,
    ):
        """
        Return sim_telarray metadata.

        Parameters
        ----------
        type: str
            Type of the configuration file (telescope, site)
        model_parameters: dict
            Model parameters dictionary.
        telescope_model_name: str
            Name of the telescope model
        additional_metadata: dict
            Dictionary with additional metadata to include using 'set'.
        telescope_design_model: str
            Name of the telescope design model.

        Returns
        -------
        list
            List with sim_telarray metadata.
        """
        meta_parameters = [
            f"config_release = {self._model_version} with simtools v{simtools.version.__version__}",
            f"config_version = {self._model_version}",
        ]
        telescope_design_model = telescope_design_model or "design_model_not_set"
        if config_type == "telescope":
            meta_parameters.extend(
                [
                    f"camera_config_name = {telescope_design_model}",
                    f"camera_config_variant = {telescope_model_name}",
                    f"camera_config_version = {self._model_version}",
                    f"optics_config_name = {telescope_design_model}",
                    f"optics_config_variant = {telescope_model_name}",
                    f"optics_config_version = {self._model_version}",
                ]
            )
            prefix = "metaparam telescope"
        elif config_type == "site":
            meta_parameters.extend(
                [
                    f"site_config_name = {self._site}",
                    "site_config_variant = ",
                    f"site_config_version = {self._model_version}",
                    f"array_config_name = {self._layout_name}",
                    "array_config_variant = ",
                    f"array_config_version = {self._model_version}",
                ]
            )
            prefix = "metaparam global"
            meta_parameters.append("metaparam global add random_seed")
        else:
            raise ValueError(f"Unknown metadata type {config_type}")

        self._add_model_parameters_to_metadata(model_parameters, meta_parameters, prefix)

        if additional_metadata:
            for key, value in additional_metadata.items():
                if value:
                    meta_parameters.append(f"{prefix} set {key}={value}")

        return meta_parameters

    def _add_model_parameters_to_metadata(self, model_parameters, meta_parameters, prefix):
        """Add model parameters to metadata."""
        if not model_parameters:
            return

        for key, value in model_parameters.items():
            simtel_name = names.get_simulation_software_name_from_parameter_name(
                key, software_name="sim_telarray", set_meta_parameter=False
            )
            if simtel_name and value.get("meta_parameter"):
                meta_parameters.append(f"{prefix} add {simtel_name}")
            simtel_name = names.get_simulation_software_name_from_parameter_name(
                key, software_name="sim_telarray", set_meta_parameter=True
            )
            if simtel_name and value.get("meta_parameter"):
                meta_parameters.append(f"{prefix} set {simtel_name}={value['value']}")

    def write_array_config_file(
        self, config_file_path, telescope_model, site_model, additional_metadata=None
    ):
        """
        Write the sim_telarray config file for an array of telescopes.

        Parameters
        ----------
        config_file_path: str or Path
            Path of the file to write on.
        telescope_model: dict of TelescopeModel
            Dictionary of TelescopeModel's instances as used by the ArrayModel instance.
        site_model: Site model
            Site model.
        additional_metadata: dict
            Dictionary with additional metadata to include.
        """
        config_file_directory = Path(config_file_path).parent
        with open(config_file_path, "w", encoding="utf-8") as file:
            self._write_header(file, "ARRAY CONFIGURATION FILE")

            file.write("#ifndef TELESCOPE\n")
            file.write("# define TELESCOPE 0\n")
            file.write("#endif\n\n")

            # TELESCOPE 0 - global parameters
            file.write("#if TELESCOPE == 0\n")
            file.write(self.TAB + "echo *****************************\n")
            file.write(self.TAB + f"echo Site: {self._site}\n")
            file.write(self.TAB + f"echo LayoutName: {self._layout_name}\n")
            file.write(self.TAB + f"echo ModelVersion: {self._model_version}\n")
            file.write(self.TAB + "echo *****************************\n\n")

            self._write_simtools_parameters(file)

            self._write_site_parameters(
                file,
                site_model.parameters,
                config_file_directory,
                telescope_model,
                additional_metadata,
            )

            file.write(self.TAB + f"maximum_telescopes = {len(telescope_model)}\n\n")

            # Default telescope in sim_telarray - 0th tel in telescope list
            _, first_telescope = next(iter(telescope_model.items()))
            invalid_telescope_name = "InvalidTelescope"
            file.write(f"# include <{invalid_telescope_name}.cfg>\n\n")
            self.write_dummy_telescope_configuration_file(
                deepcopy(first_telescope.parameters),
                config_file_directory / f"{invalid_telescope_name}.cfg",
                invalid_telescope_name,
            )

            for count, (tel_name, tel_model) in enumerate(telescope_model.items()):
                tel_config_file = tel_model.config_file_path.name
                file.write(f"% {tel_name}\n")
                file.write(f"#elif TELESCOPE == {count + 1}\n\n")
                file.write(f"# include <{tel_config_file}>\n\n")
            file.write("#endif \n\n")  # configuration files need to end with \n\n

        if additional_metadata and additional_metadata.get("random_instrument_instances"):
            self._write_random_seeds_file(additional_metadata, config_file_directory)

    def _write_random_seeds_file(self, sim_telarray_seeds, config_file_directory):
        """
        Write list of random number used to generate random instances of instrument.

        Parameters
        ----------
        random_instrument_instances: int
            Number of random instances of the instrument.
        """
        self._logger.info(
            "Writing random seed file "
            f"{config_file_directory}/{sim_telarray_seeds['seed_file_name']}"
            f" (global seed {sim_telarray_seeds['seed']})"
        )
        if sim_telarray_seeds["random_instrument_instances"] > 1024:
            raise ValueError("Number of random instances of instrument must be less than 1024")
        random_integers = sim_telarray_random_seeds(
            sim_telarray_seeds["seed"], sim_telarray_seeds["random_instrument_instances"]
        )
        with open(
            config_file_directory / sim_telarray_seeds["seed_file_name"], "w", encoding="utf-8"
        ) as file:
            file.write(
                "# Random seeds for instrument configuration generated with seed "
                f"{sim_telarray_seeds['seed']}"
                f" (model version {self._model_version}, site {self._site})\n"
            )
            for number in random_integers:
                file.write(f"{number}\n")

    def write_single_mirror_list_file(
        self, mirror_number, mirrors, single_mirror_list_file, set_focal_length_to_zero=False
    ):
        """
        Write the sim_telarray mirror list file for a single mirror.

        Parameters
        ----------
        mirror_number: int
            Mirror number.
        mirrors: Mirrors
            Instance of Mirrors.
        single_mirror_list_file: str or Path
            Path of the file to write on.
        set_focal_length_to_zero: bool
            Flag to set the focal length to zero.
        """
        (
            __,
            __,
            mirror_panel_diameter,
            focal_length,
            shape_type,
        ) = mirrors.get_single_mirror_parameters(mirror_number)
        with open(single_mirror_list_file, "w", encoding="utf-8") as file:
            self._write_header(file, "MIRROR LIST FILE", "#")

            file.write("# Column 1: X pos. [cm] (North/Down)\n")
            file.write("# Column 2: Y pos. [cm] (West/Right from camera)\n")
            file.write("# Column 3: flat-to-flat diameter [cm]\n")
            file.write(
                "# Column 4: focal length [cm], typically zero = adapting in sim_telarray.\n"
            )
            file.write(
                "# Column 5: shape type: 0=circular, 1=hex. with flat side parallel to y, "
                "2=square, 3=other hex. (default: 0)\n"
            )
            file.write(
                "# Column 6: Z pos (height above dish backplane) [cm], typ. omitted (or zero)"
                " to adapt to dish shape settings.\n"
            )
            file.write("#\n")
            file.write(
                f"0. 0. {mirror_panel_diameter.to('cm').value} "
                f"{focal_length.to('cm').value if not set_focal_length_to_zero else 0} "
                f"{shape_type} 0.\n"
            )

    def _write_header(self, file, title, comment_char="%"):
        """
        Write a generic header.

        Parameters
        ----------
        file: file
            File to write.
        title: str
            Title of the header.
        comment_char: str
            Character to be used for comments, which differs among types of config files.
        """
        header = f"{comment_char}{50 * '='}\n"
        header += f"{comment_char} {title}\n"
        header += f"{comment_char} Site: {self._site}\n"
        header += f"{comment_char} ModelVersion: {self._model_version}\n"
        header += (
            f"{comment_char} TelescopeModelName: {self._telescope_model_name}\n"
            if self._telescope_model_name is not None
            else ""
        )
        header += (
            f"{comment_char} LayoutName: {self._layout_name}\n"
            if self._layout_name is not None
            else ""
        )
        header += f"{comment_char} Label: {self._label}\n" if self._label is not None else ""
        header += f"{comment_char}{50 * '='}\n"
        header += f"{comment_char}\n"
        file.write(header)

    def _write_simtools_parameters(self, file):
        """Write simtools-specific parameters."""
        meta_items = {
            "simtools_version": simtools.version.__version__,
            "simtools_model_production_version": self._model_version,
        }
        try:
            build_opts = ascii_handler.collect_data_from_file(
                Path(self._simtel_path) / "build_opts.yml"
            )
            for key, value in build_opts.items():
                meta_items[f"simtools_{key}"] = value
        except (FileNotFoundError, TypeError):
            pass  # don't expect build_opts.yml to be present on all systems

        file.write(f"{self.TAB}% Simtools parameters\n")
        for key, value in meta_items.items():
            file.write(f"{self.TAB}metaparam global set {key} = {value}\n")

    def _write_site_parameters(
        self, file, site_parameters, model_path, telescope_model, additional_metadata=None
    ):
        """
        Write site parameters.

        Parameters
        ----------
        file: file
            File to write on.
        site_parameters: site parameters
            Site parameters.
        model_path: Path
            Path to the model for writing of additional files.
        telescope_model: dict of TelescopeModel
            Telescope models.
        additional_metadata: dict
            Dictionary with additional metadata to include.
        """
        file.write(self.TAB + "% Site parameters\n")
        for par, value in site_parameters.items():
            simtel_name, simtel_value = self._convert_model_parameters_to_simtel_format(
                names.get_simulation_software_name_from_parameter_name(
                    par, software_name="sim_telarray"
                ),
                value["value"],
                model_path,
                telescope_model,
            )
            if simtel_name is not None:
                file.write(f"{self.TAB}{simtel_name} = {simtel_value}\n")
        for meta in self._get_sim_telarray_metadata(
            "site", site_parameters, None, additional_metadata
        ):
            file.write(f"{self.TAB}{meta}\n")
        file.write("\n")

    def _convert_model_parameters_to_simtel_format(
        self, simtel_name, value, model_path, telescope_model
    ):
        """
        Convert model parameter value to simtel format.

        This might involve format or unit conversion and writing to a parameter file.

        Parameters
        ----------
        simtel_name: str
            Parameter name.
        value: any
            Value to convert.
        model_path: Path
            Path to the model for writing of additional files.
        telescope_model: dict of TelescopeModel
            Telescope models.

        Returns
        -------
        str, any
            Converted parameter name and value.
        """
        conversion_dict = {
            "array_triggers": self._write_array_triggers_file,
        }
        try:
            value = conversion_dict[simtel_name](value, model_path, telescope_model)
        except KeyError:
            pass
        except AttributeError:  # covers cases where telescope_model is None
            return None, None
        return simtel_name, value

    def _write_array_triggers_file(self, array_triggers, model_path, telescope_model):
        """
        Write array trigger definition file in simtel format.

        Parameters
        ----------
        array_triggers: dict
            Array trigger definitions.
        model_path: Path
            Path to the model for writing of additional files.
        telescope_model: dict of TelescopeModel
            Telescope models.
        """
        trigger_per_telescope_type = self._group_telescopes_by_type(telescope_model)
        hardstereo_lines, non_hardstereo_groups, all_non_hardstereo_tels, multiplicity = (
            self._process_telescope_triggers(array_triggers, trigger_per_telescope_type)
        )

        array_triggers_file = "array_triggers.dat"
        with open(model_path / array_triggers_file, "w", encoding="utf-8") as file:
            file.write("# Array trigger definition\n")
            self._write_trigger_lines(
                file, hardstereo_lines, non_hardstereo_groups, all_non_hardstereo_tels, multiplicity
            )

        return array_triggers_file

    def _group_telescopes_by_type(self, telescope_model):
        """Group telescopes by their type."""
        trigger_per_telescope_type = {}
        for count, tel_name in enumerate(telescope_model.keys()):
            telescope_type = names.get_array_element_type_from_name(tel_name)
            trigger_per_telescope_type.setdefault(telescope_type, []).append(count + 1)
        return trigger_per_telescope_type

    def _process_telescope_triggers(self, array_triggers, trigger_per_telescope_type):
        """Process telescope triggers and group them by hardstereo and parameters."""
        hardstereo_lines = []
        non_hardstereo_groups = {}
        all_non_hardstereo_tels = []
        multiplicity = None

        for tel_type, tel_list in trigger_per_telescope_type.items():
            trigger_dict = self._get_array_triggers_for_telescope_type(
                array_triggers, tel_type, len(tel_list)
            )
            width, minsep = self._extract_trigger_parameters(trigger_dict)
            multiplicity = trigger_dict["multiplicity"]["value"]  # Store for later use

            if trigger_dict.get("hard_stereo", {}).get("value"):
                line = self._build_trigger_line(
                    trigger_dict, tel_list, width, minsep, hardstereo=True
                )
                hardstereo_lines.append(line)
            else:
                key = (width, minsep)
                non_hardstereo_groups.setdefault(key, []).extend(tel_list)
                all_non_hardstereo_tels.extend(tel_list)

        return hardstereo_lines, non_hardstereo_groups, all_non_hardstereo_tels, multiplicity

    def _extract_trigger_parameters(self, trigger_dict):
        """Extract width and min_separation parameters from trigger dictionary."""
        width = trigger_dict["width"]["value"] * u.Unit(trigger_dict["width"]["unit"]).to("ns")
        minsep = None
        if all(trigger_dict["min_separation"][key] is not None for key in ["value", "unit"]):
            minsep = trigger_dict["min_separation"]["value"] * u.Unit(
                trigger_dict["min_separation"]["unit"]
            ).to("m")
        return width, minsep

    def _build_trigger_line(self, trigger_dict, tel_list, width, minsep, hardstereo=False):
        """Build a trigger line string."""
        line = f"Trigger {trigger_dict['multiplicity']['value']} of "
        line += ", ".join(map(str, tel_list))
        line += f" width {width}"
        if hardstereo:
            line += " hardstereo"
        if minsep is not None:
            line += f" minsep {minsep}"
        return line

    def _write_trigger_lines(
        self, file, hardstereo_lines, non_hardstereo_groups, all_non_hardstereo_tels, multiplicity
    ):
        """Write all trigger lines to file."""
        # Write hardstereo lines first
        for line in hardstereo_lines:
            file.write(f"{line}\n")

        # Write individual non-hardstereo groups if they have different parameters
        if len(non_hardstereo_groups) > 1:
            for (width, minsep), tel_list in non_hardstereo_groups.items():
                line = f"Trigger {multiplicity} of "
                line += ", ".join(map(str, tel_list))
                line += f" width {width}"
                if minsep is not None:
                    line += f" minsep {minsep}"
                file.write(f"{line}\n")

        # Write combined line with all non-hardstereo telescopes using shortest values
        if all_non_hardstereo_tels:
            min_width = min(width for width, minsep in non_hardstereo_groups.keys())
            min_minsep = self._get_minimum_minsep(non_hardstereo_groups)

            combined_line = f"Trigger {multiplicity} of "
            combined_line += ", ".join(map(str, sorted(all_non_hardstereo_tels)))
            combined_line += f" width {min_width}"
            if min_minsep is not None:
                combined_line += f" minsep {min_minsep}"
            file.write(f"{combined_line}\n")

    def _get_minimum_minsep(self, non_hardstereo_groups):
        """Get minimum min_separation value from groups."""
        minsep_values = [
            minsep for width, minsep in non_hardstereo_groups.keys() if minsep is not None
        ]
        return min(minsep_values) if minsep_values else None

    def _get_array_triggers_for_telescope_type(
        self, array_triggers, telescope_type, num_telescopes_of_type
    ):
        """
        Get array trigger for a specific telescope type.

        Parameters
        ----------
        array_triggers: dict
            Array trigger definitions.
        telescope_type: str
            Telescope type.
        num_telescopes_of_type: int
            Number of telescopes of the specified type.

        Returns
        -------
        dict
            Array trigger for the telescope type.
        """
        suffix = "_array"
        if num_telescopes_of_type == 1:
            suffix = "_single_telescope"
        for trigger_dict in array_triggers:
            if trigger_dict["name"] == telescope_type + suffix:
                return trigger_dict
        return None

    def write_dummy_telescope_configuration_file(
        self, parameters, config_file_path, telescope_name
    ):
        """
        Write 'dummy' telescope configuration file used as zeroth telescope in sim_telarray.

        Replaces key telescope configuration values with dummy values.

        Parameters
        ----------
        parameters: dict
            Telescope parameters used as template.
        config_file_path: str or Path
            Path of the dummy configuration file to write on.
        telescope_name: str
            Name of the telescope.
        """
        self._logger.debug(f"Writing {telescope_name} telescope config file {config_file_path}")
        dummy_defaults = {
            "camera_config_file": f"{telescope_name}_single_pixel_camera.dat",
            "discriminator_pulse_shape": f"{telescope_name}_pulse.dat",
            "fadc_pulse_shape": f"{telescope_name}_pulse.dat",
            "mirror_list": f"{telescope_name}_single_12m_mirror.dat",
            "mirror_reflectivity": f"{telescope_name}_reflectivity.dat",
            "camera_pixels": 1,
            "trigger_pixels": 1,
            "num_gains": 1,
            "fadc_bins": 10,
            "disc_bins": 10,
            "fadc_sum_bins": 10,
            "fadc_sum_offset": 0,
            "asum_threshold": 9999,
            "dsum_threshold": 9999,
            "discriminator_threshold": 9999,
            "fadc_amplitude": 1.0,
            "discriminator_amplitude": 1.0,
        }

        for key, val in dummy_defaults.items():
            if key in parameters:
                parameters[key]["value"] = val

        self.write_telescope_config_file(config_file_path, parameters, telescope_name)

        config_file_directory = Path(config_file_path).parent
        self._write_dummy_mirror_list_files(config_file_directory, telescope_name)
        self._write_dummy_camera_files(config_file_directory, telescope_name)

    def _write_dummy_mirror_list_files(self, config_directory, telescope_name):
        """Write dummy mirror list with single mirror and reflectivity file."""
        with open(
            config_directory / f"{telescope_name}_single_12m_mirror.dat", "w", encoding="utf-8"
        ) as file:
            file.write("0 0 1200 0.0 0\n")
        with open(
            config_directory / f"{telescope_name}_reflectivity.dat", "w", encoding="utf-8"
        ) as file:
            file.writelines(f"{w} 0.8\n" for w in range(200, 801, 50))

    def _write_dummy_camera_files(self, config_directory, telescope_name):
        """Write dummy camera, pulse shape, and funnels file with a single pixel."""
        with open(
            config_directory / f"{telescope_name}_single_pixel_camera.dat", "w", encoding="utf-8"
        ) as file:
            file.write(f'PixType 1   0  0 300   1 300 0.00   "{telescope_name}_funnels.dat"\n')
            file.write("Pixel 0 1 0. 0.  0  0  0 0x00 1\n")
            file.write("Trigger 1 of 0\n")

        with open(config_directory / f"{telescope_name}_funnel.dat", "w", encoding="utf-8") as file:
            file.writelines(f"{a} 0.78 1.5\n" for a in range(36))

        with open(config_directory / f"{telescope_name}_pulse.dat", "w", encoding="utf-8") as file:
            file.write("0 0 0\n")
            file.write("1 1 1\n")
            file.write("2 2 2\n")
            file.write("3 3 3\n")
            file.write("4 0 0\n")



#!/usr/bin/python3
"""Read file info and run headers from eventio (CORSIKA IACT, sim_telarray) files."""

import warnings

from eventio import EventIOFile, iact
from eventio.simtel import MCRunHeader, MCShower, RunHeader

# Suppress all UserWarnings from corsikaio - no CORSIKA versions <7.7 are supported anyway
warnings.filterwarnings("ignore", category=UserWarning, module=r"corsikaio\.subblocks\..*")


def get_corsika_run_number(file):
    """
    Return the CORSIKA run number from an eventio (CORSIKA IACT or sim_telarray) file.

    Parameters
    ----------
    file: str
        Path to the eventio file.

    Returns
    -------
    int, None
        CORSIKA run number. Returns None if not found.
    """
    run_header = get_combined_corsika_run_header(file)
    if run_header and "run" in run_header:
        return run_header["run"]
    run_header, _ = get_corsika_run_and_event_headers(file)
    try:
        return int(run_header["run_number"])
    except (TypeError, KeyError, ValueError):
        return None


def get_combined_corsika_run_header(sim_telarray_file):
    """
    Return the CORSIKA run header information from an sim_telarray file.

    Reads both RunHeader and MCRunHeader object from file and returns a merged dictionary.
    Adds primary id from the first event.

    Parameters
    ----------
    sim_telarray_file: str
        Path to the sim_telarray file.

    Returns
    -------
    dict, None
        CORSIKA run header. Returns None if not found.
    """
    run_header = mc_run_header = None
    primary_id = None

    with EventIOFile(sim_telarray_file) as f:
        for o in f:
            if isinstance(o, RunHeader) and run_header is None:
                run_header = o.parse()
            elif isinstance(o, MCRunHeader) and mc_run_header is None:
                mc_run_header = o.parse()
            elif isinstance(o, MCShower):  # get primary_id from first MCShower
                primary_id = o.parse().get("primary_id")
            if run_header and mc_run_header and primary_id is not None:
                break

    run_header = run_header or {}
    mc_run_header = mc_run_header or {}
    if primary_id is not None:
        mc_run_header["primary_id"] = primary_id
    return run_header | mc_run_header or None


def get_corsika_run_and_event_headers(corsika_iact_file):
    """
    Return the CORSIKA run and event headers from a CORSIKA IACT eventio file.

    Parameters
    ----------
    corsika_iact_file: str, Path
        Path to the CORSIKA IACT eventio file.

    Returns
    -------
    tuple
        CORSIKA run header and event header as dictionaries.
    """
    run_header = event_header = None

    with EventIOFile(corsika_iact_file) as f:
        for o in f:
            if isinstance(o, iact.RunHeader) and run_header is None:
                run_header = o.parse()
            elif isinstance(o, iact.EventHeader) and event_header is None:
                event_header = o.parse()
            if run_header and event_header:
                break

    return run_header, event_header


def get_simulated_events(event_io_file):
    """
    Return the number of shower and MC events from a simulation (eventio) file.

    For a sim_telarray file, the number of simulated showers and MC events is
    determined by counting the number of MCShower (type id 2020) and MCEvent
    objects (type id 2021). For a CORSIKA IACT file, the number of simulated
    showers is determined by counting the number of IACTShower (type id 1202).

    Parameters
    ----------
    event_io_file: str, Path
        Path to the eventio file.

    Returns
    -------
    tuple
        Number of showers and number of MC events (MC events for sim_telarray files only).
    """
    counts = {1202: 0, 2020: 0, 2021: 0}
    with EventIOFile(event_io_file) as f:
        for o in f:
            t = o.header.type
            if t in counts:
                counts[t] += 1
    return counts[2020] if counts[2020] else counts[1202], counts[2021]



#!/usr/bin/python3
"""Functions for plotting mirror panel layout information."""

import logging
from pathlib import Path

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection

from simtools.io import io_handler
from simtools.model.mirrors import Mirrors
from simtools.model.telescope_model import TelescopeModel
from simtools.visualization import visualize

logger = logging.getLogger(__name__)

PATCH_STYLE = {"alpha": 0.8, "edgecolor": "black", "facecolor": "dodgerblue"}
LABEL_STYLE = {"ha": "center", "va": "center", "fontsize": 10, "color": "white", "weight": "bold"}
STATS_BOX_STYLE = {"boxstyle": "round", "facecolor": "wheat", "alpha": 0.8}
OBSERVER_TEXT = "for an observer facing the mirrors"


def _rotate_coordinates_clockwise_90(x_pos, y_pos):
    """
    Rotate coordinates 90 degrees clockwise for observer-facing view.

    Plots are rotated by 90 degrees clockwise to present the point of view of a
    person standing on the camera platform.

    Important: Any transformation applied to one mirror list must be applied consistently to
    all (primary) mirror lists. Inconsistent transformations would result in incorrect mirror
    configurations when running sim_telarray simulations.
    """
    return y_pos, -x_pos


def _detect_segmentation_type(data_file_path):
    """
    Detect the type of segmentation file (ring, shape, or standard).

    Parameters
    ----------
    data_file_path : Path
        Path to the segmentation data file

    Returns
    -------
    str
        One of "ring", "shape", or "standard"
    """
    with open(data_file_path, encoding="utf-8") as f:
        for line in f:
            line_lower = line.strip().lower()
            if line_lower.startswith("#") or not line_lower:
                continue
            if line_lower.startswith("ring"):
                return "ring"
            if line_lower.startswith(("hex", "yhex")):
                return "shape"
    return "standard"


def plot(config, output_file, db_config=None):
    """
    Plot mirror panel layout based on configuration.

    Parameters
    ----------
    config : dict
        Configuration dictionary containing:
        - parameter: str, parameter name (e.g., "mirror_list", "primary_mirror_segmentation")
        - site: str, site name (e.g., "North", "South")
        - telescope: str, telescope name (e.g., "LSTN-01")
        - parameter_version: str, optional, parameter version
        - model_version: str, optional, model version
        - title: str, optional, plot title
    output_file : str or Path
        Path where to save the plot (without extension)
    db_config : dict, optional
        Database configuration dictionary

    Returns
    -------
    None
        The function saves the plot to the specified output file.
    """
    tel_model = TelescopeModel(
        site=config["site"],
        telescope_name=config["telescope"],
        model_version=config.get("model_version"),
        db_config=db_config,
        ignore_software_version=True,
    )

    output_path = io_handler.IOHandler().get_output_directory()

    parameter_name = config["parameter"]
    parameter_value = tel_model.get_parameter_value(parameter_name)
    tel_model.export_model_files(destination_path=output_path)

    mirror_file = parameter_value
    data_file_path = Path(output_path / mirror_file)

    parameter_type = config["parameter"]

    if parameter_type in ("primary_mirror_segmentation", "secondary_mirror_segmentation"):
        segmentation_type = _detect_segmentation_type(data_file_path)

        if segmentation_type == "ring":
            fig = plot_mirror_ring_segmentation(
                data_file_path=data_file_path,
                telescope_model_name=config["telescope"],
                parameter_type=parameter_type,
            )
        elif segmentation_type == "shape":
            fig = plot_mirror_shape_segmentation(
                data_file_path=data_file_path,
                telescope_model_name=config["telescope"],
                parameter_type=parameter_type,
            )
        else:
            fig = plot_mirror_segmentation(
                data_file_path=data_file_path,
                telescope_model_name=config["telescope"],
                parameter_type=parameter_type,
            )
    else:
        mirrors = Mirrors(mirror_list_file=data_file_path)
        fig = plot_mirror_layout(
            mirrors=mirrors,
            mirror_file_path=data_file_path,
            telescope_model_name=config["telescope"],
        )

    visualize.save_figure(fig, output_file)
    plt.close(fig)


def plot_mirror_layout(mirrors, mirror_file_path, telescope_model_name):
    """
    Plot the mirror panel layout from a Mirrors object.

    Parameters
    ----------
    mirrors : Mirrors
        Mirrors object containing mirror panel data including positions,
        diameters, focal lengths, and shape types
    mirror_file_path : Path or str
        Path to the mirror list file
    telescope_model_name : str
        Name of the telescope model (e.g., "LSTN-01", "MSTN-01")

    Returns
    -------
    matplotlib.figure.Figure
        The generated figure object
    """
    logger.info(f"Plotting mirror layout for {telescope_model_name}")

    fig, ax = plt.subplots(figsize=(10, 10))

    x_pos = mirrors.mirror_table["mirror_x"].to("cm").value
    y_pos = mirrors.mirror_table["mirror_y"].to("cm").value
    x_pos, y_pos = _rotate_coordinates_clockwise_90(x_pos, y_pos)
    diameter = mirrors.mirror_diameter.to("cm").value
    shape_type = mirrors.shape_type
    focal_lengths = mirrors.mirror_table["focal_length"].to("cm").value

    mirror_ids = (
        mirrors.mirror_table["mirror_panel_id"]
        if "mirror_panel_id" in mirrors.mirror_table.colnames
        else list(range(len(x_pos)))
    )

    # MST mirrors are numbered from 1 at the bottom to N at the top
    if telescope_model_name and "MST" in telescope_model_name.upper():
        n_mirrors = len(mirror_ids)
        mirror_ids = [n_mirrors - mid for mid in mirror_ids]

    patches, colors = _create_mirror_patches(x_pos, y_pos, diameter, shape_type, focal_lengths)

    collection = PatchCollection(
        patches,
        cmap="viridis",
        edgecolor="black",
        linewidth=0.5,
    )
    collection.set_array(np.array(colors))
    ax.add_collection(collection)

    _add_mirror_labels(ax, x_pos, y_pos, mirror_ids, max_labels=20)

    _configure_mirror_plot(ax, x_pos, y_pos)

    cbar = plt.colorbar(collection, ax=ax, pad=0.02)
    cbar.set_label("Focal length [cm]", fontsize=14)

    _add_mirror_statistics(ax, mirrors, mirror_file_path, x_pos, y_pos, diameter)

    return fig


def plot_mirror_segmentation(data_file_path, telescope_model_name, parameter_type):
    """
    Plot mirror segmentation layout from a segmentation file.

    Parameters
    ----------
    data_file_path : Path or str
        Path to the segmentation data file containing mirror segment positions,
        diameters, and shape types in standard numeric format
    telescope_model_name : str
        Name of the telescope model (e.g., "LSTN-01", "MSTN-01")
    parameter_type : str
        Type of segmentation parameter (e.g., "primary_mirror_segmentation",
        "secondary_mirror_segmentation")

    Returns
    -------
    matplotlib.figure.Figure
        The generated figure object
    """
    logger.info(f"Plotting {parameter_type} for {telescope_model_name}")

    segmentation_data = _read_segmentation_file(data_file_path)

    fig, ax = plt.subplots(figsize=(10, 10))

    x_pos = segmentation_data["x"]
    y_pos = segmentation_data["y"]
    x_pos, y_pos = _rotate_coordinates_clockwise_90(x_pos, y_pos)
    diameter = segmentation_data["diameter"]
    shape_type = segmentation_data["shape_type"]
    segment_ids = segmentation_data["segment_ids"]

    patches, colors = _create_mirror_patches(x_pos, y_pos, diameter, shape_type, segment_ids)

    collection = PatchCollection(
        patches,
        cmap="tab20",
        edgecolor="black",
        linewidth=0.8,
    )
    collection.set_array(np.array(colors))
    ax.add_collection(collection)

    _add_mirror_labels(ax, x_pos, y_pos, segment_ids, max_labels=30)

    _configure_mirror_plot(ax, x_pos, y_pos)

    cbar = plt.colorbar(collection, ax=ax, pad=0.02)
    cbar.set_label("Segment ID", fontsize=14)

    n_segments = len(set(segment_ids))
    stats_text = (
        f"Number of segments: {len(x_pos)}\n"
        f"Number of segment groups: {n_segments}\n"
        f"Segment diameter: {diameter:.1f} cm"
    )

    ax.text(
        0.02,
        0.98,
        stats_text,
        transform=ax.transAxes,
        fontsize=11,
        verticalalignment="top",
        bbox=STATS_BOX_STYLE,
    )

    return fig


def _create_mirror_patches(x_pos, y_pos, diameter, shape_type, color_values):
    """Create matplotlib patches for mirror panels or segments."""
    patches = [
        _create_single_mirror_patch(x, y, diameter, shape_type) for x, y in zip(x_pos, y_pos)
    ]
    return patches, list(color_values)


def _read_segmentation_file(data_file_path):
    """Read mirror segmentation file and extract segment information."""
    x_pos = []
    y_pos = []
    diameter = None
    shape_type = None
    segment_ids = []

    with open(data_file_path, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            parts = line.split()
            if len(parts) < 5:
                continue

            try:
                x_pos.append(float(parts[0]))
                y_pos.append(float(parts[1]))
            except ValueError:
                continue

            diameter = _extract_diameter(parts, diameter)
            shape_type = _extract_shape_type(parts, shape_type)
            segment_ids.append(_extract_segment_id(parts, len(segment_ids)))

    if len(x_pos) == 0:
        logger.warning(f"No valid numeric data found in segmentation file: {data_file_path}")

    return {
        "x": np.array(x_pos),
        "y": np.array(y_pos),
        "diameter": diameter if diameter is not None else 150.0,
        "shape_type": shape_type if shape_type is not None else 3,
        "segment_ids": segment_ids,
    }


def _extract_diameter(parts, current_diameter):
    """Extract diameter from parts or return current value."""
    return float(parts[2]) if current_diameter is None else current_diameter


def _extract_shape_type(parts, current_shape_type):
    """Extract shape type from parts or return current value."""
    return int(parts[4]) if current_shape_type is None else current_shape_type


def _extract_segment_id(parts, default_id):
    """Extract segment ID from parts or return default."""
    if len(parts) >= 8:
        seg_id_str = parts[7].split("=")[-1] if "=" in parts[7] else parts[7]
        return int("".join(filter(str.isdigit, seg_id_str)))
    return default_id


def _create_single_mirror_patch(x, y, diameter, shape_type):
    """Create a single matplotlib patch for a mirror panel (hexagonal only)."""
    base_orientation = 0 if shape_type == 1 else np.pi / 2
    # Rotate hexagon counter-clockwise to compensate for clockwise coordinate rotation
    orientation = base_orientation - np.pi / 2

    return mpatches.RegularPolygon(
        (x, y),
        numVertices=6,
        radius=diameter / np.sqrt(3),
        orientation=orientation,
    )


def _add_mirror_labels(ax, x_pos, y_pos, mirror_ids, max_labels=20):
    """Add mirror panel ID labels to the plot."""
    mirror_data = sorted(zip(mirror_ids, x_pos, y_pos), key=lambda item: item[0])

    for i, (mid, x, y) in enumerate(mirror_data):
        if i < max_labels:
            ax.text(
                x,
                y,
                str(mid),
                ha="center",
                va="center",
                fontsize=6,
                color="white",
                weight="bold",
            )


def _configure_mirror_plot(ax, x_pos, y_pos):
    """Add titles, labels, and limits."""
    ax.set_aspect("equal")

    if len(x_pos) == 0 or len(y_pos) == 0:
        logger.warning("No valid mirror data found for plotting")
        ax.set_xlim(-1000, 1000)
        ax.set_ylim(-1000, 1000)
        ax.text(0, 0, "No valid mirror data", ha="center", va="center", fontsize=14, color="red")
        return

    x_min, x_max = np.min(x_pos), np.max(x_pos)
    y_min, y_max = np.min(y_pos), np.max(y_pos)

    x_padding = (x_max - x_min) * 0.15
    y_padding = (y_max - y_min) * 0.15

    ax.set_xlim(x_min - x_padding, x_max + x_padding)
    ax.set_ylim(y_min - y_padding, y_max + y_padding)

    plt.xlabel("X position [cm]", fontsize=14)
    plt.ylabel("Y position [cm]", fontsize=14)
    plt.grid(True, alpha=0.3)
    plt.tick_params(axis="both", which="major", labelsize=12)

    ax.text(
        0.02,
        0.02,
        OBSERVER_TEXT,
        transform=ax.transAxes,
        fontsize=10,
        ha="left",
        va="bottom",
        style="italic",
        color="black",
    )


def _extract_float_after_keyword(line, keyword):
    """Extract first float value after keyword in line."""
    if keyword not in line:
        return None
    try:
        part = line.split(keyword, 1)[1] if keyword == "=" else line.split(keyword)[-1]
        return float(part.strip().split()[0])
    except (ValueError, IndexError):
        return None


def _read_mirror_file_metadata(mirror_file_path):
    """Read metadata from mirror .dat file header (Rmax and total surface area)."""
    metadata = {}
    patterns = [
        (("Total surface area:", "Total mirror surface area:"), ":", "total_surface_area"),
        (("Rmax =", "Rmax="), "=", "rmax"),
        (("mirrors are inside a radius of",), "of", "rmax"),
    ]

    try:
        with open(mirror_file_path, encoding="utf-8") as f:
            for line in f:
                if not line.startswith("#"):
                    break
                for triggers, keyword, key in patterns:
                    if key not in metadata and any(t in line for t in triggers):
                        if (val := _extract_float_after_keyword(line, keyword)) is not None:
                            metadata[key] = val
                            break
    except OSError as e:
        logger.warning(f"Could not read mirror file metadata: {e}")

    return metadata


def _add_mirror_statistics(ax, mirrors, mirror_file_path, x_pos, y_pos, diameter):
    """Add mirror statistics text to the plot."""
    n_mirrors = mirrors.number_of_mirrors

    metadata = _read_mirror_file_metadata(mirror_file_path)
    max_radius = metadata.get("rmax")
    total_area = metadata.get("total_surface_area")

    # Calculate values if not available from file
    if max_radius is None:
        max_radius = np.sqrt(np.max(x_pos**2 + y_pos**2)) / 100.0

    if total_area is None:
        panel_area = 3 * np.sqrt(3) / 2 * (diameter / 200.0) ** 2
        total_area = n_mirrors * panel_area

    stats_text = (
        f"Number of mirrors: {n_mirrors}\n"
        f"Mirror diameter: {diameter:.1f} cm\n"
        f"Max radius: {max_radius:.2f} m\n"
        f"Total surface area: {total_area:.2f} $m^{2}$"
    )

    ax.text(
        0.02,
        0.98,
        stats_text,
        transform=ax.transAxes,
        fontsize=11,
        verticalalignment="top",
        bbox=STATS_BOX_STYLE,
    )


def _read_ring_segmentation_data(data_file_path):
    """Read ring segmentation data from file."""
    rings = []

    with open(data_file_path, encoding="utf-8") as f:
        for line in f:
            if not line.startswith("#") and not line.startswith("%"):
                if line.lower().startswith("ring"):
                    parts = line.split()
                    rings.append(
                        {
                            "nseg": int(parts[1].strip()),
                            "rmin": float(parts[2].strip()),
                            "rmax": float(parts[3].strip()),
                            "dphi": float(parts[4].strip()),
                            "phi0": float(parts[5].strip()),
                        }
                    )

    return rings


def _plot_single_ring(ax, ring, cmap, color_index):
    """Plot a single ring with its segments."""
    rmin, rmax = ring["rmin"], ring["rmax"]
    nseg, phi0 = ring["nseg"], ring["phi0"]
    dphi = ring["dphi"]

    # Angular gap between segments (in degrees) - represents the physical gaps
    angular_gap = 0.3  # degrees

    if nseg > 1:
        dphi_rad = dphi * np.pi / 180
        phi0_rad = phi0 * np.pi / 180
        gap_rad = angular_gap * np.pi / 180

        for i in range(nseg):
            theta_i = i * dphi_rad + phi0_rad

            # Fill segment with small gap on each side
            n_theta = 100
            theta_seg = np.linspace(theta_i + gap_rad, theta_i + dphi_rad - gap_rad, n_theta)

            color_value = (color_index + i % 10) / 20.0  # Normalize to 0-1
            color = cmap(color_value)
            ax.fill_between(theta_seg, rmin, rmax, color=color, alpha=0.8)


def _add_ring_radius_label(ax, angle, radius, label_text):
    """Add a radius label at the specified angle and radius."""
    ax.text(
        angle,
        radius,
        label_text,
        ha="center",
        va="center",
        fontsize=9,
        color="red",
        weight="bold",
        bbox={"boxstyle": "round,pad=0.3", "facecolor": "white", "alpha": 0.8},
    )


def plot_mirror_ring_segmentation(data_file_path, telescope_model_name, parameter_type):
    """
    Plot mirror ring segmentation layout.

    Parameters
    ----------
    data_file_path : Path or str
        Path to the segmentation data file containing ring definitions with format:
        ring <nseg> <rmin> <rmax> <dphi> <phi0>
    telescope_model_name : str
        Name of the telescope model (e.g., "LSTN-01", "MSTN-01")
    parameter_type : str
        Type of segmentation parameter (e.g., "primary_mirror_segmentation",
        "secondary_mirror_segmentation")

    Returns
    -------
    matplotlib.figure.Figure or None
        The generated figure object, or None if no ring data found
    """
    logger.info(f"Plotting ring {parameter_type} for {telescope_model_name}")

    rings = _read_ring_segmentation_data(data_file_path)

    if not rings:
        logger.warning(f"No ring data found in {data_file_path}")
        return None

    fig, ax = plt.subplots(subplot_kw={"projection": "polar"}, figsize=(10, 10))

    cmap = mcolors.LinearSegmentedColormap.from_list("mirror_blue", ["#deebf7", "#3182bd"])

    for i, ring in enumerate(rings):
        _plot_single_ring(ax, ring, cmap, color_index=i * 10)

    max_radius = max(ring["rmax"] for ring in rings)
    label_padding = max_radius * 0.04
    ax.set_ylim([0, max_radius + label_padding])
    ax.set_yticklabels([])
    ax.set_rgrids([])
    ax.spines["polar"].set_visible(False)

    label_angle = 30 * np.pi / 180

    for ring in rings:
        theta_full = np.linspace(0, 2 * np.pi, 360)
        ax.plot(
            theta_full,
            np.repeat(ring["rmin"], len(theta_full)),
            ":",
            color="gray",
            lw=0.8,
            alpha=0.5,
        )
        ax.plot(
            theta_full,
            np.repeat(ring["rmax"], len(theta_full)),
            ":",
            color="gray",
            lw=0.8,
            alpha=0.5,
        )

        _add_ring_radius_label(ax, label_angle, ring["rmin"], f"{ring['rmin']:.3f}")
        _add_ring_radius_label(ax, label_angle, ring["rmax"], f"{ring['rmax']:.3f}")

    ax.text(
        label_angle,
        max_radius + label_padding * 2.5,
        "[cm]",
        ha="center",
        va="center",
        fontsize=10,
        weight="bold",
        color="red",
    )

    if len(rings) == 2:
        stats_text = (
            f"Inner ring segments: {rings[0]['nseg']}\nOuter ring segments: {rings[1]['nseg']}"
        )
    else:
        stats_lines = [f"Ring {i + 1} segments: {ring['nseg']}" for i, ring in enumerate(rings)]
        stats_text = "\n".join(stats_lines)

    ax.text(
        0.02,
        0.98,
        stats_text,
        transform=ax.transAxes,
        fontsize=11,
        verticalalignment="top",
        bbox=STATS_BOX_STYLE,
    )

    ax.text(
        0.02,
        0.02,
        OBSERVER_TEXT,
        transform=ax.transAxes,
        fontsize=10,
        ha="left",
        va="bottom",
        style="italic",
        color="black",
    )

    plt.tight_layout()

    return fig


def _parse_segment_id_line(line_stripped):
    """Extract segment ID from a line if it contains segment ID information."""
    try:
        return int(line_stripped.split()[-1])
    except (ValueError, IndexError):
        return 0


def _is_skippable_line(line_stripped):
    """Check if line should be skipped (empty or comment)."""
    return not line_stripped or line_stripped.startswith(("#", "%"))


def _parse_shape_line(line_stripped, shape_segments, segment_ids, current_segment_id):
    """Parse and append a single shape segmentation line."""
    entries = line_stripped.split()

    if any(line_stripped.lower().startswith(s) for s in ["hex", "yhex"]) and len(entries) >= 5:
        shape_segments.append(
            {
                "shape": entries[0].lower(),
                "x": float(entries[2]),
                "y": float(entries[3]),
                "diameter": float(entries[4]),
                "rotation": float(entries[5]) if len(entries) > 5 else 0.0,
            }
        )
        segment_ids.append(current_segment_id if current_segment_id > 0 else len(shape_segments))


def _read_shape_segmentation_file(data_file_path):
    """
    Read shape segmentation file.

    Parameters
    ----------
    data_file_path : Path
        Path to segmentation file

    Returns
    -------
    tuple
        (shape_segments, segment_ids)
    """
    shape_segments, segment_ids = [], []
    current_segment_id = 0

    with open(data_file_path, encoding="utf-8") as f:
        for line in f:
            line_stripped = line.strip()

            if "segment id" in line_stripped.lower():
                current_segment_id = _parse_segment_id_line(line_stripped)
            elif not _is_skippable_line(line_stripped):
                _parse_shape_line(line_stripped, shape_segments, segment_ids, current_segment_id)

    return shape_segments, segment_ids


def _add_segment_label(ax, x, y, label):
    """Add a label at the specified position."""
    ax.text(x, y, str(label), **LABEL_STYLE)


def _create_shape_patches(ax, shape_segments, segment_ids):
    """
    Create patches for shape segments (hexagons).

    Parameters
    ----------
    shape_segments : list
        List of shape segment dictionaries
    segment_ids : list
        List of segment IDs
    ax : matplotlib.axes.Axes
        Axes to add text labels to

    Returns
    -------
    tuple
        (patches, maximum_radius)
    """
    patches, maximum_radius = [], 0

    for i_seg, seg in enumerate(shape_segments):
        x, y, diam, rot = seg["x"], seg["y"], seg["diameter"], seg["rotation"]
        maximum_radius = max(maximum_radius, abs(x) + diam / 2, abs(y) + diam / 2)

        patch = mpatches.RegularPolygon(
            (x, y),
            numVertices=6,
            radius=diam / np.sqrt(3),
            orientation=np.deg2rad(rot),
            **PATCH_STYLE,
        )

        patches.append(patch)
        label = segment_ids[i_seg] if i_seg < len(segment_ids) else i_seg + 1
        _add_segment_label(ax, x, y, label)

    return patches, maximum_radius


def plot_mirror_shape_segmentation(data_file_path, telescope_model_name, parameter_type):
    """
    Plot mirror shape segmentation layout.

    Parameters
    ----------
    data_file_path : Path or str
        Path to the segmentation data file containing explicit shape definitions
        (hex, yhex) with positions, diameters, and rotations
    telescope_model_name : str
        Name of the telescope model (e.g., "LSTN-01", "MSTN-design")
    parameter_type : str
        Type of segmentation parameter (e.g., "primary_mirror_segmentation",
        "secondary_mirror_segmentation")

    Returns
    -------
    matplotlib.figure.Figure
        The generated figure object
    """
    logger.info(f"Plotting shape {parameter_type} for {telescope_model_name}")

    shape_segments, segment_ids = _read_shape_segmentation_file(data_file_path)

    for seg in shape_segments:
        seg["x"], seg["y"] = _rotate_coordinates_clockwise_90(seg["x"], seg["y"])
        seg["rotation"] = seg["rotation"] - 90

    fig, ax = plt.subplots(figsize=(10, 10))

    # Create patches for shape segments
    all_patches, maximum_radius = _create_shape_patches(ax, shape_segments, segment_ids)

    collection = PatchCollection(all_patches, match_original=True)
    ax.add_collection(collection)

    ax.set_aspect("equal")
    padding = maximum_radius * 0.1 if maximum_radius > 0 else 100
    ax.set_xlim(-maximum_radius - padding, maximum_radius + padding)
    ax.set_ylim(-maximum_radius - padding, maximum_radius + padding)

    plt.xlabel("X position [cm]", fontsize=14)
    plt.ylabel("Y position [cm]", fontsize=14)
    plt.grid(True, alpha=0.3)
    plt.tick_params(axis="both", which="major", labelsize=12)

    ax.text(
        0.02,
        0.02,
        OBSERVER_TEXT,
        transform=ax.transAxes,
        fontsize=10,
        ha="left",
        va="bottom",
        style="italic",
        color="black",
    )

    total_segments = len(shape_segments)
    if segment_ids and total_segments > 0:
        stats_text = f"Number of segments: {len(set(segment_ids))}"
    elif total_segments > 0:
        stats_text = f"Number of segments: {total_segments}"
    else:
        stats_text = "No segment data"

    ax.text(
        0.02,
        0.98,
        stats_text,
        transform=ax.transAxes,
        fontsize=11,
        verticalalignment="top",
        bbox=STATS_BOX_STYLE,
    )

    return fig



"""
Optical PSF plotting functions for parameter optimization visualization.

This module provides plotting functionality for PSF parameter optimization,
including parameter comparison plots, convergence plots, and PSF diameter vs off-axis plots.
"""

import logging

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages

from simtools.ray_tracing.ray_tracing import RayTracing
from simtools.visualization import visualize

# Constants
RADIUS = "Radius [cm]"
CUMULATIVE_PSF = "Cumulative PSF"
MAX_OFFSET_DEFAULT = 4.5  # Maximum off-axis angle in degrees
OFFSET_STEPS_DEFAULT = 0.1  # Step size for off-axis angle sampling
DEFAULT_FRACTION = 0.8  # Default PSF containment fraction

logger = logging.getLogger(__name__)


def get_psf_diameter_label(fraction, unit="cm"):
    """
    Generate PSF diameter label based on containment fraction.

    Parameters
    ----------
    fraction : float
        PSF containment fraction (e.g., 0.8 for D80, 0.95 for D95)
    unit : str, optional
        Unit for the label (default: "cm")

    Returns
    -------
    str
        Formatted PSF diameter label (e.g., "D80 (cm)", "D95 (cm)", "D95")
    """
    percentage = int(fraction * 100)
    if unit:
        return f"D{percentage} ({unit})"
    return f"D{percentage}"


def get_significance_label(p_value):
    """Get significance label for p-value."""
    if p_value > 0.05:  # null hypothesis not rejected at the 95% level
        return "GOOD"
    if p_value > 0.01:  # null hypothesis rejected at 95% but not at 99% level
        return "FAIR"
    return "POOR"  # null hypothesis rejected at 99% level


def _format_metric_text(
    psf_diameter,
    metric,
    fraction=DEFAULT_FRACTION,
    p_value=None,
    use_ks_statistic=False,
    second_metric=None,
):
    """
    Format metric text for display in plots.

    Parameters
    ----------
    psf_diameter : float
        PSF diameter value
    metric : float
        Primary metric value (RMSD or KS statistic)
    fraction : float, optional
        PSF containment fraction (default: 0.8)
    p_value : float, optional
        P-value from KS test
    use_ks_statistic : bool
        If True, metric is KS statistic; if False, metric is RMSD
    second_metric : float, optional
        Second metric value to display alongside the primary metric

    Returns
    -------
    str
        Formatted metric text
    """
    psf_label = get_psf_diameter_label(fraction, unit="")
    psf_text = f"{psf_label} = {psf_diameter:.5f} cm"

    # Create metric text based on the optimization method
    if second_metric is not None:
        # Special case: show both RMSD and KS statistic (for final best plot)
        metric_text = f"RMSD = {metric:.4f}"  # metric is RMSD in this case
        metric_text += f"\nKS statistic = {second_metric:.4f}"  # second_metric is KS statistic
        if p_value is not None:
            metric_text += f"\np-value = {p_value:.4f}"
    elif use_ks_statistic:
        metric_text = f"KS stat = {metric:.6f}"
        if p_value is not None:
            significance = get_significance_label(p_value)
            metric_text += f"\np-value = {p_value:.6f} ({significance})"
    else:
        metric_text = f"RMSD = {metric:.4f}"

    return f"{psf_text}\n{metric_text}"


def _create_base_plot_figure(data_to_plot, simulated_data=None):
    """
    Create base figure for PSF parameter plots.

    Parameters
    ----------
    data_to_plot : dict
        Data dictionary for plotting
    simulated_data : array, optional
        Simulated data to add to the plot

    Returns
    -------
    tuple
        (fig, ax) - figure and axis objects
    """
    plot_data = data_to_plot.copy()

    if simulated_data is not None:
        plot_data["simulated"] = simulated_data

    fig = visualize.plot_1d(
        plot_data,
        plot_difference=True,
        no_markers=True,
    )
    ax = fig.get_axes()[0]
    ax.set_ylim(0, 1.05)
    ax.set_ylabel(CUMULATIVE_PSF)
    return fig, ax


def _build_parameter_title(pars, is_best):
    """Build parameter title string for plots."""
    title_prefix = "* " if is_best else ""
    return (
        f"{title_prefix}reflection = "
        f"{pars['mirror_reflection_random_angle'][0]:.5f}, "
        f"{pars['mirror_reflection_random_angle'][1]:.5f}, "
        f"{pars['mirror_reflection_random_angle'][2]:.5f}\n"
        f"align_vertical = {pars['mirror_align_random_vertical'][0]:.5f}, "
        f"{pars['mirror_align_random_vertical'][1]:.5f}, "
        f"{pars['mirror_align_random_vertical'][2]:.5f}, "
        f"{pars['mirror_align_random_vertical'][3]:.5f}\n"
        f"align_horizontal = {pars['mirror_align_random_horizontal'][0]:.5f}, "
        f"{pars['mirror_align_random_horizontal'][1]:.5f}, "
        f"{pars['mirror_align_random_horizontal'][2]:.5f}, "
        f"{pars['mirror_align_random_horizontal'][3]:.5f}"
    )


def _add_metric_text_box(ax, metrics_text, is_best):
    """Add metric text box to plot."""
    psf_color = "red" if is_best else "black"
    psf_weight = "bold" if is_best else "normal"

    ax.text(
        0.5,
        0.3,
        metrics_text,
        verticalalignment="center",
        horizontalalignment="left",
        transform=ax.transAxes,
        color=psf_color,
        weight=psf_weight,
        bbox={"boxstyle": "round,pad=0.3", "facecolor": "yellow", "alpha": 0.7}
        if is_best
        else None,
    )


def _add_plot_annotations(
    ax,
    fig,
    pars,
    psf_diameter,
    metric,
    is_best,
    fraction=DEFAULT_FRACTION,
    p_value=None,
    use_ks_statistic=False,
    second_metric=None,
):
    """
    Add title, text annotations, and best parameter indicators to plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The plot axes
    fig : matplotlib.figure.Figure
        The plot figure
    pars : dict
        Parameter set dictionary
    psf_diameter : float
        PSF diameter value
    metric : float
        Primary metric value
    is_best : bool
        Whether this is the best parameter set
    fraction : float, optional
        PSF containment fraction (default: 0.8)
    p_value : float, optional
        P-value from KS test
    use_ks_statistic : bool
        If True, metric is KS statistic; if False, metric is RMSD
    second_metric : float, optional
        Second metric value to display
    """
    title = _build_parameter_title(pars, is_best)
    ax.set_title(title)

    metrics_text = _format_metric_text(
        psf_diameter, metric, fraction, p_value, use_ks_statistic, second_metric
    )
    _add_metric_text_box(ax, metrics_text, is_best)

    if is_best:
        fig.text(
            0.02,
            0.02,
            "* Best parameter set (lowest RMSD)",
            fontsize=8,
            style="italic",
            color="red",
        )


def create_psf_parameter_plot(
    data_to_plot,
    pars,
    psf_diameter,
    metric,
    is_best,
    pdf_pages,
    fraction=DEFAULT_FRACTION,
    p_value=None,
    use_ks_statistic=False,
    second_metric=None,
):
    """
    Create a plot for PSF simulation results.

    Parameters
    ----------
    data_to_plot : dict
        Data dictionary for plotting.
    pars : dict
        Parameter set dictionary.
    psf_diameter : float
        PSF diameter value.
    metric : float
        RMSD value (if use_ks_statistic=False) or KS statistic (if use_ks_statistic=True).
    is_best : bool
        Whether this is the best parameter set.
    pdf_pages : PdfPages
        PDF pages object for saving plots.
    fraction : float, optional
        PSF containment fraction (default: 0.8).
    p_value : float, optional
        P-value from KS test (only used when use_ks_statistic=True).
    use_ks_statistic : bool, optional
        If True, metric is KS statistic; if False, metric is RMSD.
    second_metric : float, optional
        Second metric value to display alongside the primary metric (for final best plot).
    """
    fig, ax = _create_base_plot_figure(data_to_plot)

    _add_plot_annotations(
        ax,
        fig,
        pars,
        psf_diameter,
        metric,
        is_best,
        fraction,
        p_value,
        use_ks_statistic,
        second_metric,
    )

    pdf_pages.savefig(fig, bbox_inches="tight")
    plt.clf()


def create_detailed_parameter_plot(
    pars,
    ks_statistic,
    psf_diameter,
    simulated_data,
    data_to_plot,
    is_best,
    pdf_pages,
    fraction=DEFAULT_FRACTION,
    p_value=None,
):
    """
    Create a detailed plot for a parameter set showing all parameter values.

    Parameters
    ----------
    pars : dict
        Parameter set dictionary
    ks_statistic : float
        KS statistic value for this parameter set
    psf_diameter : float
        PSF diameter value for this parameter set
    simulated_data : array
        Simulated data for plotting
    data_to_plot : dict
        Data dictionary for plotting
    is_best : bool
        Whether this is the best parameter set
    pdf_pages : PdfPages
        PDF pages object to save the plot
    fraction : float, optional
        PSF containment fraction (default: 0.8)
    p_value : float, optional
        P-value from KS test for statistical significance
    """
    # Check if we have valid simulated data for plotting
    if simulated_data is None:
        logger.warning(
            "No simulated data available for plotting this parameter set, skipping plot creation"
        )
        return

    try:
        fig, ax = _create_base_plot_figure(data_to_plot, simulated_data)
    except (ValueError, RuntimeError, KeyError, TypeError) as e:
        logger.error(f"Failed to create plot for parameters: {e}")
        return

    _add_plot_annotations(
        ax,
        fig,
        pars,
        psf_diameter,
        ks_statistic,
        is_best,
        fraction,
        p_value,
        use_ks_statistic=True,
        second_metric=None,
    )

    pdf_pages.savefig(fig, bbox_inches="tight")
    plt.clf()


def create_parameter_progression_plots(
    results, best_pars, data_to_plot, pdf_pages, fraction=DEFAULT_FRACTION
):
    """
    Create plots for all parameter sets showing optimization progression.

    Parameters
    ----------
    results : list
        List of (pars, ks_statistic, p_value, psf_diameter, simulated_data) tuples
    best_pars : dict
        Best parameter set for highlighting
    data_to_plot : dict
        Data dictionary for plotting
    pdf_pages : PdfPages
        PDF pages object to save plots
    fraction : float, optional
        PSF containment fraction (default: 0.8)
    """
    logger.info("Creating plots for all parameter sets...")

    for i, (pars, ks_statistic, p_value, psf_diameter, simulated_data) in enumerate(results):
        if simulated_data is None:
            logger.warning(f"No simulated data for iteration {i}, skipping plot")
            continue

        is_best = pars is best_pars
        logger.info(f"Creating plot {i + 1}/{len(results)}{' (BEST)' if is_best else ''}")

        create_detailed_parameter_plot(
            pars,
            ks_statistic,
            psf_diameter,
            simulated_data,
            data_to_plot,
            is_best,
            pdf_pages,
            fraction,
            p_value,
        )


def create_gradient_descent_convergence_plot(
    gd_results, threshold, output_file, fraction=DEFAULT_FRACTION, use_ks_statistic=False
):
    """
    Create convergence plot during gradient descent.

    Parameters
    ----------
    gd_results : list
        List of (params, metric, p_value, psf_diameter, simulated_data) tuples from gradient descent
    threshold : float
        Optimization metric threshold used for convergence
    output_file : Path
        Output file path for saving the plot
    fraction : float, optional
        PSF containment fraction for labeling (default: 0.8)
    use_ks_statistic : bool
        Whether to use KS statistic or RMSD labels and titles
    """
    logger.info("Creating gradient descent convergence plot...")

    # Check if results include p-values (for KS statistic mode)
    has_p_values = len(gd_results[0]) >= 4 and gd_results[0][2] is not None

    if has_p_values and use_ks_statistic:
        _, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(10, 12), tight_layout=True)
    else:
        _, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 8), tight_layout=True)

    iterations = list(range(len(gd_results)))
    metrics = [r[1] for r in gd_results]
    psf_diameters = [r[3] for r in gd_results]

    metric_name = "KS Statistic" if use_ks_statistic else "RMSD"
    psf_label = get_psf_diameter_label(fraction)

    # Plot optimization metric progression
    ax1.plot(iterations, metrics, "b.-", linewidth=2, markersize=6)
    ax1.axhline(y=threshold, color="r", linestyle="--", label=f"Threshold: {threshold}")
    ax1.set_xlabel("Iteration")
    ax1.set_ylabel(metric_name)
    ax1.set_title(f"Gradient Descent Convergence - {metric_name}")
    ax1.grid(True, alpha=0.3)
    ax1.legend()

    # Plot PSF diameter progression
    ax2.plot(iterations, psf_diameters, "g.-", linewidth=2, markersize=6)
    ax2.set_xlabel("Iteration")
    ax2.set_ylabel(psf_label)
    ax2.set_title(f"Gradient Descent Convergence - {get_psf_diameter_label(fraction, unit='')}")
    ax2.grid(True, alpha=0.3)

    # Plot p-value progression if available and using KS statistic
    if has_p_values and use_ks_statistic:
        p_values = [r[2] for r in gd_results if r[2] is not None]
        p_iterations = [i for i, r in enumerate(gd_results) if r[2] is not None]

        ax3.plot(p_iterations, p_values, "m.-", linewidth=2, markersize=6)
        ax3.axhline(
            y=0.05, color="orange", linestyle="--", alpha=0.7, label="p = 0.05 (significance)"
        )
        ax3.axhline(
            y=0.01, color="r", linestyle="--", alpha=0.7, label="p = 0.01 (high significance)"
        )
        ax3.set_xlabel("Iteration")
        ax3.set_ylabel("p-value")
        ax3.set_title("Gradient Descent Convergence - Statistical Significance")
        ax3.set_yscale("log")  # Log scale for p-values
        ax3.grid(True, alpha=0.3)
        ax3.legend()

    plt.savefig(output_file, bbox_inches="tight")
    plt.close()

    logger.info(f"Convergence plot saved to {output_file}")


def create_monte_carlo_uncertainty_plot(
    mc_results, output_file, fraction=DEFAULT_FRACTION, use_ks_statistic=False
):
    """
    Create uncertainty analysis plots showing optimization metric and p-value distributions.

    Parameters
    ----------
    mc_results : tuple
        Results from Monte Carlo analysis: (mean_metric, std_metric, metric_values,
        mean_p, std_p, p_values, mean_psf_diameter, std_psf_diameter, psf_diameter_values)
    output_file : Path
        Output file path for saving the plot
    fraction : float, optional
        PSF containment fraction for labeling (default: 0.8)
    use_ks_statistic : bool, optional
        Whether KS statistic mode is being used (affects filename suffix)
    """
    (
        mean_metric,
        std_metric,
        metric_values,
        mean_p_value,
        _,  # std_p_value (unused)
        p_values,
        mean_psf_diameter,
        std_psf_diameter,
        psf_diameter_values,
    ) = mc_results

    logger.info("Creating Monte Carlo uncertainty analysis plot...")

    # Check if we have valid p-values to determine if this is KS statistic or RMSD mode
    valid_p_values = [p for p in p_values if p is not None] if p_values else []
    is_ks_mode = len(valid_p_values) > 0

    # Create subplot layout based on mode
    if is_ks_mode:
        # KS mode: 2x2 layout with all 4 plots
        _, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 10), tight_layout=True)
    else:
        # RMSD mode: 1x2 layout with only metric and PSF diameter plots
        _, (ax1, ax3) = plt.subplots(1, 2, figsize=(12, 5), tight_layout=True)

    # Metric histogram (KS statistic or RMSD)
    metric_name = "KS Statistic" if is_ks_mode else "RMSD"
    psf_label = get_psf_diameter_label(fraction)

    ax1.hist(metric_values, bins=20, alpha=0.7, color="blue", edgecolor="black")
    ax1.axvline(
        mean_metric, color="red", linestyle="--", linewidth=2, label=f"Mean: {mean_metric:.6f}"
    )
    ax1.axvline(
        mean_metric - std_metric,
        color="orange",
        linestyle=":",
        alpha=0.7,
        label=f"$\\sigma$: {std_metric:.6f}",
    )
    ax1.axvline(mean_metric + std_metric, color="orange", linestyle=":", alpha=0.7)
    ax1.set_xlabel(metric_name)
    ax1.set_ylabel("Counts")
    ax1.set_title(f"{metric_name} Distribution")
    ax1.legend()
    ax1.grid(True, alpha=0.3)

    # p-value histogram (only for KS statistic mode)
    if is_ks_mode:
        ax2.hist(valid_p_values, bins=20, alpha=0.7, color="magenta", edgecolor="black")
        if mean_p_value is not None:
            ax2.axvline(
                mean_p_value,
                color="red",
                linestyle="--",
                linewidth=2,
                label=f"Mean: {mean_p_value:.6f}",
            )
        ax2.axvline(0.05, color="orange", linestyle="--", alpha=0.7, label="p = 0.05")
        ax2.axvline(0.01, color="red", linestyle="--", alpha=0.7, label="p = 0.01")
        ax2.set_xlabel("p-value")
        ax2.set_ylabel("Counts")
        ax2.set_title("p-value Distribution")
        ax2.legend()
        ax2.grid(True, alpha=0.3)

    # PSF diameter histogram
    ax3.hist(psf_diameter_values, bins=20, alpha=0.7, color="green", edgecolor="black")
    ax3.axvline(
        mean_psf_diameter,
        color="red",
        linestyle="--",
        linewidth=2,
        label=f"Mean: {mean_psf_diameter:.4f} cm",
    )
    ax3.axvline(
        mean_psf_diameter - std_psf_diameter,
        color="orange",
        linestyle=":",
        alpha=0.7,
        label=f"$\\sigma$: {std_psf_diameter:.4f} cm",
    )
    ax3.axvline(mean_psf_diameter + std_psf_diameter, color="orange", linestyle=":", alpha=0.7)
    ax3.set_xlabel(psf_label)
    ax3.set_ylabel("Counts")
    ax3.set_title(f"{get_psf_diameter_label(fraction, unit='')} Distribution")
    ax3.legend()
    ax3.grid(True, alpha=0.3)

    # Scatter plot: Metric vs p-value (only for KS statistic mode)
    if is_ks_mode:
        ax4.scatter(metric_values, valid_p_values, alpha=0.6, color="purple")
        ax4.axhline(y=0.05, color="orange", linestyle="--", alpha=0.7, label="p = 0.05")
        ax4.axhline(y=0.01, color="red", linestyle="--", alpha=0.7, label="p = 0.01")
        ax4.set_xlabel(metric_name)
        ax4.set_ylabel("p-value")
        ax4.set_title(f"{metric_name} vs p-value Correlation")
        ax4.set_yscale("log")
        ax4.legend()
        ax4.grid(True, alpha=0.3)

    # Save plot in both PDF and PNG formats with appropriate suffix
    suffix = "_ks" if use_ks_statistic else "_rmsd"

    # Generate base filename without extension
    base_path = output_file.with_suffix("")
    base_name = str(base_path)

    # Add suffix and save in both formats
    pdf_file = f"{base_name}{suffix}.pdf"
    png_file = f"{base_name}{suffix}.png"

    plt.savefig(pdf_file, bbox_inches="tight")
    plt.savefig(png_file, bbox_inches="tight", dpi=150)
    plt.close()

    logger.info(f"Monte Carlo uncertainty plot saved to {pdf_file}")
    logger.info(f"Monte Carlo uncertainty plot saved to {png_file}")


def create_psf_vs_offaxis_plot(tel_model, site_model, args_dict, best_pars, output_dir):
    """
    Create PSF diameter vs off-axis angle plot using the best parameters.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object.
    site_model : SiteModel
        Site model object.
    args_dict : dict
        Dictionary containing parsed command-line arguments.
    best_pars : dict
        Best parameter set.
    output_dir : Path
        Output directory for saving plots.
    """
    fraction = args_dict.get("fraction", DEFAULT_FRACTION)
    psf_label_cm = get_psf_diameter_label(fraction, unit="cm")
    psf_label_deg = get_psf_diameter_label(fraction, unit="degrees")

    logger.info(f"Creating {psf_label_cm} vs off-axis angle plot with best parameters...")

    # Apply best parameters to telescope model
    tel_model.overwrite_parameters(best_pars)

    # Create off-axis angle array
    max_offset = args_dict.get("max_offset", MAX_OFFSET_DEFAULT)
    offset_steps = args_dict.get("offset_steps", OFFSET_STEPS_DEFAULT)
    off_axis_angles = np.linspace(
        0,
        max_offset,
        int(max_offset / offset_steps) + 1,
    )

    ray = RayTracing(
        telescope_model=tel_model,
        site_model=site_model,
        simtel_path=args_dict["simtel_path"],
        zenith_angle=args_dict["zenith"] * u.deg,
        source_distance=args_dict["src_distance"] * u.km,
        off_axis_angle=off_axis_angles * u.deg,
    )

    logger.info(f"Running ray tracing for {len(off_axis_angles)} off-axis angles...")
    ray.simulate(test=args_dict.get("test", False), force=True)
    ray.analyze(force=True)

    for key in ["d80_cm", "d80_deg"]:
        plt.figure(figsize=(10, 6), tight_layout=True)

        ray.plot(key, marker="o", linestyle="-", color="blue", linewidth=2, markersize=6)

        parameters_text = (
            f"Best Parameters: \n"
            f"reflection=["
            f"{', '.join(f'{x:.4f}' for x in best_pars['mirror_reflection_random_angle'])}],\n"
            f"align_horizontal=["
            f"{', '.join(f'{x:.4f}' for x in best_pars['mirror_align_random_horizontal'])}]\n"
            f"align_vertical=["
            f"{', '.join(f'{x:.4f}' for x in best_pars['mirror_align_random_vertical'])}]\n"
        )
        plt.title(parameters_text)
        plt.xlabel("Off-axis Angle (degrees)")
        plt.ylabel(psf_label_cm if "_cm" in key else psf_label_deg)
        plt.ylim(bottom=0)
        plt.xticks(rotation=45)
        plt.xlim(0, max_offset)
        plt.grid(True, alpha=0.3)

        # Create dynamic file name based on fraction
        psf_identifier = get_psf_diameter_label(fraction, unit="").lower()  # e.g., "d80", "d95"
        unit_suffix = "cm" if "_cm" in key else "deg"
        plot_file_name = f"{tel_model.name}_best_params_{psf_identifier}_{unit_suffix}.png"
        plot_file = output_dir.joinpath(plot_file_name)
        visualize.save_figure(
            plt, plot_file, figure_format=["png"], log_title=f"{psf_label_cm} vs off-axis ({key})"
        )

    plt.close("all")


def setup_pdf_plotting(args_dict, output_dir, tel_model_name):
    """
    Set up PDF plotting for gradient descent optimization if requested.

    Parameters
    ----------
    args_dict : dict
        Dictionary containing command-line arguments with plot_all flag.
    output_dir : Path
        Directory where the PDF file will be saved.
    tel_model_name : str
        Name of the telescope model for filename generation.

    Returns
    -------
    PdfPages or None
        PdfPages object for saving plots if plotting is requested, None otherwise.

    Notes
    -----
    Creates a PDF file for saving cumulative PSF plots during gradient descent
    optimization. Returns None if plotting is not requested via the plot_all flag.
    """
    if not args_dict.get("plot_all", False):
        return None
    pdf_filename = output_dir / f"psf_gradient_descent_plots_{tel_model_name}.pdf"
    pdf_pages = PdfPages(pdf_filename)
    logger.info(f"Creating cumulative PSF plots for each iteration (saving to {pdf_filename})")
    return pdf_pages


def create_optimization_plots(args_dict, gd_results, tel_model, data_to_plot, output_dir):
    """
    Create optimization plots for all iterations if requested.

    Parameters
    ----------
    args_dict : dict
        Dictionary containing command-line arguments with save_plots flag.
    gd_results : list
        List of (params, rmsd, _, psf_diameter, _) tuples from gradient descent optimization.
    tel_model : TelescopeModel
        Telescope model object for naming files.
    data_to_plot : dict
        Dictionary containing measured PSF data.
    output_dir : Path
        Directory where the PDF file will be saved.

    Notes
    -----
    Creates a PDF file with plots for optimization iterations. Only creates plots
    every 5 iterations plus the final iteration to avoid excessively large files.
    Returns early if save_plots flag is not set.
    """
    if not args_dict.get("save_plots", False):
        return

    fraction = args_dict.get("fraction", DEFAULT_FRACTION)
    pdf_filename = output_dir.joinpath(f"psf_optimization_results_{tel_model.name}.pdf")
    pdf_pages = PdfPages(pdf_filename)
    logger.info(f"Creating PSF plots for each optimization iteration (saving to {pdf_filename})")

    for i, (params, rmsd, _, psf_diameter, _) in enumerate(gd_results):
        if i % 5 == 0 or i == len(gd_results) - 1:
            create_psf_parameter_plot(
                data_to_plot,
                params,
                psf_diameter,
                rmsd,
                is_best=(i == len(gd_results) - 1),
                pdf_pages=pdf_pages,
                fraction=fraction,
                use_ks_statistic=False,
            )
    pdf_pages.close()



"""Handle interpolation between multiple StatisticalUncertaintyEvaluator instances."""

import logging

import astropy.units as u
import numpy as np
from scipy.interpolate import griddata

from simtools.production_configuration.derive_production_statistics import (
    ProductionStatisticsDerivator,
)


class InterpolationHandler:
    """
    Calculate the required events for production via interpolation from a grid.

    This class provides methods to interpolate production statistics across a grid of
    parameter values (azimuth, zenith, NSB, offset) and energy.
    """

    def __init__(self, evaluators, metrics: dict, grid_points_production: list):
        """
        Initialize the InterpolationHandler.

        Parameters
        ----------
        evaluators : list
            List of StatisticalUncertaintyEvaluator instances.
        metrics : dict
            Dictionary of metrics to use for production statistics.
        grid_points_production : list
            List of grid points for interpolation, each being a dictionary with keys
            'azimuth', 'zenith_angle', 'nsb', 'offset' etc.
        """
        self._logger = logging.getLogger(__name__)
        self.evaluators = evaluators
        self.metrics = metrics
        self.grid_points_production = grid_points_production

        self._initialize_derivators()
        self._extract_grid_properties()

        self.data, self.grid_points = self._build_data_array()
        self.interpolated_production_statistics = None
        self.interpolated_production_statistics_with_energy = None
        self._non_flat_mask = None

    def _initialize_derivators(self):
        """Initialize production statistics derivators for all evaluators."""
        self.derive_production_statistics = [
            ProductionStatisticsDerivator(e, self.metrics) for e in self.evaluators
        ]

        self.production_statistics = [
            derivator.derive_statistics(return_sum=False)
            for derivator in self.derive_production_statistics
        ]
        self.production_statistics_sum = [
            derivator.derive_statistics(return_sum=True)
            for derivator in self.derive_production_statistics
        ]

    def _extract_grid_properties(self):
        """Extract grid properties from evaluators."""
        self.azimuths = [e.grid_point[1].to(u.deg).value for e in self.evaluators]
        self.zeniths = [e.grid_point[2].to(u.deg).value for e in self.evaluators]
        self.nsbs = [e.grid_point[3] for e in self.evaluators]
        self.offsets = [e.grid_point[4].to(u.deg).value for e in self.evaluators]
        self.energy_grids = [
            (e.data["bin_edges_low"][:-1] + e.data["bin_edges_high"][:-1]) / 2
            for e in self.evaluators
        ]
        self.energy_thresholds = np.array([e.energy_threshold for e in self.evaluators])

        # Check if energy grids are consistent
        if self.energy_grids and not all(
            np.array_equal(self.energy_grids[0], grid) for grid in self.energy_grids
        ):
            self._logger.warning(
                "Energy grids are not identical across evaluators. "
                "Using the first evaluator's energy grid for interpolation."
            )

    def _build_data_array(self):
        """
        Build a data array with interpolated values across all dimensions including energy.

        Returns
        -------
        np.ndarray
            The data array with interpolated values.
        np.ndarray
            The corresponding grid points.
        """
        if not self.evaluators:
            return np.array([]), np.array([])

        flat_data_list = []
        flat_grid_points = []

        for i, (energy_grid, production_statistics) in enumerate(
            zip(self.energy_grids, self.production_statistics)
        ):
            az = self.azimuths[i]
            zen = self.zeniths[i]
            nsb = self.nsbs[i]
            offset = self.offsets[i]

            az_array = np.full(len(energy_grid), az)
            zen_array = np.full(len(energy_grid), zen)
            nsb_array = np.full(len(energy_grid), nsb)
            offset_array = np.full(len(energy_grid), offset)

            grid_points = np.column_stack(
                [energy_grid.to(u.TeV).value, az_array, zen_array, nsb_array, offset_array]
            )

            flat_grid_points.append(grid_points)
            flat_data_list.append(production_statistics)

        flat_grid_points = np.vstack(flat_grid_points)
        flat_data = np.hstack(flat_data_list)

        sorted_indices = np.argsort(flat_grid_points[:, 0])
        sorted_grid_points = flat_grid_points[sorted_indices]
        sorted_data = flat_data[sorted_indices]

        return sorted_data, sorted_grid_points

    def _remove_flat_dimensions(self, grid_points, threshold=1e-6):
        """
        Identify and remove flat dimensions (dimensions with no variance).

        Parameters
        ----------
        grid_points : np.ndarray
            Grid points to analyze.
        threshold : float, optional
            Threshold for determining flatness, by default 1e-6

        Returns
        -------
        tuple
            (reduced_grid_points, non_flat_mask)
        """
        if grid_points.size == 0:
            return grid_points, np.array([], dtype=bool)

        variance = np.var(grid_points, axis=0)
        non_flat_mask = variance > threshold

        if not np.any(non_flat_mask):
            self._logger.warning(
                "All dimensions are flat. Keeping all dimensions for interpolation."
            )
            return grid_points, np.ones_like(variance, dtype=bool)

        reduced_grid_points = grid_points[:, non_flat_mask]
        return reduced_grid_points, non_flat_mask

    def build_grid_points_no_energy(self):
        """
        Build grid points without energy dimension.

        Returns
        -------
        tuple
            (production_statistics, grid_points_no_energy)
        """
        if not self.evaluators:
            self._logger.error("No evaluators available for grid point building.")
            return np.array([]), np.array([])

        flat_data_list = []
        flat_grid_points = []

        for i, production_statistics_sum in enumerate(self.production_statistics_sum):
            az = self.azimuths[i]
            zen = self.zeniths[i]
            nsb = self.nsbs[i]
            offset = self.offsets[i]

            flat_data_list.append(float(production_statistics_sum.value))

            grid_point = np.array([[az, zen, nsb, offset]])
            flat_grid_points.append(grid_point)

        flat_grid_points = np.vstack(flat_grid_points)
        return flat_data_list, flat_grid_points

    def _prepare_energy_independent_data(self):
        """
        Prepare data for energy-independent interpolation.

        Returns
        -------
        tuple
            (production_statistic, grid_points_no_energy)
        """
        production_statistic, grid_points_no_energy = self.build_grid_points_no_energy()
        production_statistic = np.array(production_statistic, dtype=float)
        grid_points_no_energy, non_flat_mask = self._remove_flat_dimensions(grid_points_no_energy)

        self._non_flat_mask = non_flat_mask  # Store for later use
        return production_statistic, grid_points_no_energy

    def _prepare_production_grid_points(self):
        """
        Convert grid_points_production to a format suitable for interpolation.

        Returns
        -------
        np.ndarray
            Reduced production grid points.
        """
        production_grid_points = []

        for point in self.grid_points_production:
            production_grid_points.append(
                [
                    point["azimuth"]["value"],
                    point["zenith_angle"]["value"],
                    point["nsb"]["value"],
                    point["offset"]["value"],
                ]
            )

        production_grid_points = np.array(production_grid_points)

        return production_grid_points[:, self._non_flat_mask]

    def _perform_interpolation(self, grid_points, values, query_points, method="linear"):
        """
        Perform interpolation using griddata.

        Parameters
        ----------
        grid_points : np.ndarray
            Grid points for interpolation.
        values : np.ndarray
            Values at the grid points.
        query_points : np.ndarray
            Query points for interpolation.
        method : str, optional
            Interpolation method, by default "linear".

        Returns
        -------
        np.ndarray
            Interpolated values.
        """
        self._logger.debug(f"Grid points shape: {grid_points.shape}")
        self._logger.debug(f"Values shape: {values.shape}")
        self._logger.debug(f"Query points shape: {query_points.shape}")

        return griddata(
            grid_points,
            values,
            query_points,
            method=method,
            fill_value=np.nan,
            rescale=True,
        )

    def _perform_interpolation_with_energy(self):
        """
        Perform energy-dependent interpolation.

        Returns
        -------
        np.ndarray
            Energy-dependent interpolated values.
        """
        # Get grid points with energy dimension
        grid_points_energy = self.grid_points
        grid_points_energy, _ = self._remove_flat_dimensions(grid_points_energy)

        # Build energy query grid
        reduced_production_grid_points = self._prepare_production_grid_points()
        energy_grid = self.energy_grids[0] if self.energy_grids else []

        energy_query_grid = []
        for energy in energy_grid:
            for grid_point in reduced_production_grid_points:
                energy_query_grid.append(np.hstack([energy.to(u.TeV).value, grid_point]))

        energy_query_grid = np.array(energy_query_grid)

        self._logger.debug(f"Grid points with energy shape: {grid_points_energy.shape}")
        self._logger.debug(f"Data shape: {self.data.shape}")
        self._logger.debug(f"Energy query grid shape: {energy_query_grid.shape}")

        interpolated_values = self._perform_interpolation(
            grid_points_energy, self.data, energy_query_grid
        )

        reshaped = interpolated_values.reshape(
            len(reduced_production_grid_points), len(energy_grid)
        )
        return np.array([reshaped])

    def interpolate(self) -> np.ndarray:
        """
        Interpolate production statistics at the grid points specified in grid_points_production.

        This method performs two types of interpolation:
        1. Energy-independent interpolation using the sum of production statistics
        2. Energy-dependent interpolation for each energy bin

        Returns
        -------
        np.ndarray
            Interpolated values at the query points.
        """
        if not self.evaluators:
            self._logger.error("No evaluators available for interpolation.")
            return np.array([])

        # Energy-independent interpolation
        production_statistic, grid_points_no_energy = self._prepare_energy_independent_data()
        reduced_production_grid_points = self._prepare_production_grid_points()

        self.interpolated_production_statistics = self._perform_interpolation(
            grid_points_no_energy, production_statistic, reduced_production_grid_points
        )

        # Energy-dependent interpolation
        self.interpolated_production_statistics_with_energy = (
            self._perform_interpolation_with_energy()
        )

        return self.interpolated_production_statistics

    def plot_comparison(self, grid_point_index=0):
        """
        Plot a comparison between interpolated production statistics and reconstructed events.

        Parameters
        ----------
        grid_point_index : int, optional
            Index of the grid point to plot, by default 0

        Returns
        -------
        matplotlib.axes.Axes
            The Axes object containing the plot.
        """
        import matplotlib.pyplot as plt  # pylint: disable=C0415

        if not self.evaluators:
            self._logger.error("No evaluators available for plotting.")
            _, ax = plt.subplots()
            ax.text(0.5, 0.5, "No data available", ha="center", va="center")
            return ax

        # Use first evaluator for energy bins
        bin_edges_low = self.evaluators[0].data["bin_edges_low"][:-1]
        bin_edges_high = self.evaluators[0].data["bin_edges_high"][:-1]
        midpoints = (bin_edges_low + bin_edges_high) / 2

        if (
            self.interpolated_production_statistics_with_energy is None
            or len(self.interpolated_production_statistics_with_energy) == 0
            or len(self.interpolated_production_statistics_with_energy[0]) <= grid_point_index
        ):
            self._logger.warning(
                f"Invalid grid point index {grid_point_index}. Using index 0 instead."
            )
            grid_point_index = 0

        _, ax = plt.subplots()

        if (
            self.interpolated_production_statistics_with_energy is not None
            and len(self.interpolated_production_statistics_with_energy) > 0
        ):
            interpolated_stats = self.interpolated_production_statistics_with_energy[0][
                grid_point_index
            ]
            ax.plot(midpoints, interpolated_stats, label="Interpolated Production Statistics")

        reconstructed_event_histogram, _ = np.histogram(
            self.evaluators[0].data["event_energies_reco"],
            bins=self.evaluators[0].data["bin_edges_low"],
        )
        ax.plot(midpoints, reconstructed_event_histogram, label="Reconstructed Events")

        ax.legend()
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.set_xlabel("Energy (TeV)")
        ax.set_ylabel("Event Count")
        ax.set_title("Comparison of Interpolated and Reconstructed Events")

        return ax



#!/usr/bin/python3

r"""Produces a markdown file for calibration reports."""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.reporting.docs_auto_report_generator import ReportGenerator


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=("Produce a markdown report for calibration parameters."),
    )

    config.parser.add_argument(
        "--all_model_versions",
        action="store_true",
        help="Produce reports for all model versions.",
    )

    return config.initialize(
        db_config=True,
        simulation_model=["model_version"],
    )


def main():
    """Produce a markdown file for calibration reports."""
    app_context = startup_application(_parse)

    output_path = app_context.io_handler.get_output_directory()

    generator = ReportGenerator(
        db_config=app_context.db_config, args=app_context.args, output_path=output_path
    )
    generator.auto_generate_calibration_reports()

    if app_context.args.get("all_model_versions"):
        app_context.logger.info("Calibration reports for all model versions produced successfully.")
    else:
        app_context.logger.info(
            f"Calibration reports for model version {app_context.args.get('model_version')}"
            " produced successfully."
        )
    app_context.logger.info(f"Output path: {output_path}")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
r"""
Plot tabular data for a single model parameter using default plotting configurations.

Uses plotting configurations as defined in the model parameters schema files.

Command line arguments
----------------------
parameter (str, required)
    Model parameter to plot (e.g., 'atmospheric_profile').
parameter_version (str, required)
    Version of the model parameter to plot (e.g., '1.0.0').
site (str, required)
    Site for which the model parameter is defined (e.g., 'North').
telescope (str, optional)
    Telescope for which the model parameter is defined (e.g., 'LSTN-01').
output_file (str, required)
    Output file name (without suffix).
plot_type (str, optional)
    Type of plot as defined in the schema file.
    Use '--plot_type all' to plot all types defined in the schema.

Example
-------

Plot tabular data for a specific type defined in the schema file:

.. code-block:: console

    simtools-plot-tabular-data-for-model-parameter \\
        --parameter atmospheric_profile \\
        --parameter_version 1.0.0 \\
        --site North \\
        --plot_type refractive_index_vs_altitude

Plot tabular data for all types defined in the schema file:

.. code-block:: console

    simtools-plot-tabular-data-for-model-parameter \\
        --parameter fadc_pulse_shape
        --parameter_version 1.0.0 \\
        --site North \\
        --telescope LSTN-01 \\
        --plot_type all

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.visualization import plot_tables


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Plots tabular data for a model parameter.",
    )

    config.parser.add_argument("--parameter", type=str, required=True, help="Parameter name.")
    config.parser.add_argument(
        "--plot_type",
        help="Plot type as defined in the schema file.",
        type=str,
        required=True,
        default=None,
    )
    return config.initialize(db_config=True, simulation_model=["telescope", "parameter_version"])


def main():
    """Plot tabular data."""
    app_context = startup_application(_parse)

    plot_configs, output_files = plot_tables.generate_plot_configurations(
        parameter=app_context.args["parameter"],
        parameter_version=app_context.args["parameter_version"],
        site=app_context.args["site"],
        telescope=app_context.args.get("telescope"),
        output_path=app_context.io_handler.get_output_directory(),
        plot_type=app_context.args["plot_type"],
        db_config=app_context.db_config,
    )

    for plot_config, output_file in zip(plot_configs, output_files):
        plot_tables.plot(
            config=plot_config,
            output_file=output_file,
            db_config=app_context.db_config,
        )
        MetadataCollector.dump(app_context.args, output_file=output_file, add_activity_name=True)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
r"""
    Validate a file or files in a directory using a schema.

    Input files can be metadata, schema, or data files in yaml, json, or ecsv format.
    For model parameters, the schema files are taken from the simtools model parameter
    schema directory by default.

    Command line arguments
    ----------------------
    file_name (str)
      input file to be validated
    file_directory (str)
        directory with json files of model parameters to be validated
    schema (str)
      schema file (jsonschema format) used for validation
    data_type (str)
        type of input data (allowed types: metadata, schema, data, model_parameter)

    Example
    -------

    Validate metadata of a file:

    .. code-block:: console

        simtools-validate-file-using-schema \\
         --file_name tests/resources/MLTdata-preproduction.meta.yml \\
         --schema simtools/schemas/metadata.metaschema.yml \\
         --data_type metadata

    Validate schema of a file:

    .. code-block:: console

        simtools-validate-file-using-schema \\
         --file_name tests/resources/model_parameters/schema-0.3.0/num_gains-1.0.0.json \\
         --schema src/simtools/schemas/model_parameter.metaschema.yml \\
         --data_type schema

    Validate all model parameter files in a directory:

    .. code-block:: console

        simtools-validate-file-using-schema \\
         --file_directory tests/resources/model_parameters \\
         --data_type model_parameter

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.constants import MODEL_PARAMETER_SCHEMA_PATH
from simtools.data_model import metadata_collector, schema, validate_data


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Validate a file (metadata, schema, or data file) using a schema.",
    )
    config.parser.add_argument(
        "--file_name",
        help="File to be validated (full path or name pattern, e.g., '*.json')",
        default="*.json",
    )
    config.parser.add_argument(
        "--file_directory",
        help=(
            "Directory with files to be validated. "
            "If no schema file is provided, the assumption is that model "
            "parameters are validated and the schema files are taken from "
            f"{MODEL_PARAMETER_SCHEMA_PATH}."
        ),
    )
    config.parser.add_argument("--schema", help="Schema file", required=False)
    config.parser.add_argument(
        "--data_type",
        help="Type of input data",
        choices=["metadata", "schema", "data", "model_parameter"],
        default="data",
    )
    config.parser.add_argument(
        "--check_exact_data_type",
        help="Require exact data type for validation",
        action="store_true",
    )
    config.parser.add_argument(
        "--ignore_software_version",
        help="Ignore software version check.",
        action="store_true",
    )
    return config.initialize(paths=False)


def main():
    """Validate a file or files in a directory using a schema."""
    app_context = startup_application(_parse)

    file_name = app_context.args.get("file_name")
    file_directory = app_context.args.get("file_directory")
    schema_file = app_context.args.get("schema")
    data_type = app_context.args.get("data_type").lower()

    if data_type == "metadata":
        # metadata_collector runs the metadata validation by default, no need to do anything else
        metadata_collector.MetadataCollector(None, metadata_file_name=file_name)
        app_context.logger.info(f"Successful validation of metadata {file_name}")

    elif data_type == "schema":
        schema.validate_schema_from_files(
            file_directory=file_directory,
            file_name=file_name,
            schema_file=schema_file,
            ignore_software_version=app_context.args.get("ignore_software_version", False),
        )
    else:
        validate_data.DataValidator.validate_data_files(
            file_name=file_name,
            file_directory=file_directory,
            is_model_parameter=(data_type == "model_parameter"),
            check_exact_data_type=app_context.args.get("check_exact_data_type", False),
            schema_file=schema_file,
        )


if __name__ == "__main__":
    main()



"""Simulation runner for camera efficiency calculations."""

import logging
from pathlib import Path

from simtools.io import ascii_handler
from simtools.runners.simtel_runner import SimtelRunner
from simtools.utils import general


class SimulatorCameraEfficiency(SimtelRunner):
    """
    Interface with the testeff tool of sim_telarray to perform camera efficiency simulations.

    Parameters
    ----------
    telescope_model: TelescopeModel
        Instance of TelescopeModel class.
    site_model: SiteModel
        Instance of SiteModel class.
    label: str
        Instance label. Important for output file naming.
    simtel_path: str or Path
        Location of sim_telarray installation.
    file_simtel: str or Path
        Location of the sim_telarray testeff tool output file.
    zenith_angle: float
        Zenith angle given in the config to CameraEfficiency.
    nsb_spectrum: str or Path
        Path to the nsb spectrum file.
    skip_correction_to_nsb_spectrum: bool
        If True, skip the correction to the original altitude where the NSB spectrum was derived.
    """

    def __init__(
        self,
        telescope_model,
        site_model,
        label=None,
        simtel_path=None,
        file_simtel=None,
        file_log=None,
        zenith_angle=None,
        nsb_spectrum=None,
        skip_correction_to_nsb_spectrum=False,
    ):
        """Initialize SimtelRunner."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimulatorCameraEfficiency")

        super().__init__(label=label, simtel_path=simtel_path)

        self._telescope_model = telescope_model
        self._site_model = site_model
        self.label = label if label is not None else self._telescope_model.label

        self._file_simtel = file_simtel
        self._file_log = file_log
        self.zenith_angle = zenith_angle
        self.nsb_spectrum = nsb_spectrum
        self.skip_correction_to_nsb_spectrum = skip_correction_to_nsb_spectrum

    @property
    def nsb_spectrum(self):
        """nsb_spectrum property."""
        return self._nsb_spectrum

    @nsb_spectrum.setter
    def nsb_spectrum(self, nsb_spectrum):
        """Setter for nsb_spectrum."""
        if nsb_spectrum is not None:
            self._nsb_spectrum = self._validate_or_fix_nsb_spectrum_file_format(nsb_spectrum)
        else:
            self._nsb_spectrum = (
                self._site_model.config_file_directory
                / Path(self._site_model.get_parameter_value("nsb_reference_spectrum")).name
            )

    def _make_run_command(self, run_number=None, input_file=None):  # pylint: disable=unused-argument
        """Prepare the command used to run testeff."""
        self._logger.debug("Preparing the command to run testeff")

        pixel_shape = self._telescope_model.camera.get_pixel_shape()
        pixel_shape_cmd = "-hpix" if pixel_shape in [1, 3] else "-spix"
        pixel_diameter = self._telescope_model.camera.get_pixel_diameter()

        focal_length = self._telescope_model.get_telescope_effective_focal_length("m", True)

        mirror_class = self._telescope_model.get_parameter_value("mirror_class")
        curvature_radius = self._get_curvature_radius(mirror_class)
        camera_transmission = self._telescope_model.get_parameter_value("camera_transmission")

        camera_filter_file = self._telescope_model.get_parameter_value("camera_filter")
        # testeff does not support 2D distributions
        if self._telescope_model.is_file_2d("camera_filter"):
            camera_filter_file = self._get_one_dim_distribution(
                "camera_filter", "camera_filter_incidence_angle"
            )

        mirror_reflectivity = self._telescope_model.get_parameter_value("mirror_reflectivity")
        if mirror_class == 2:
            mirror_reflectivity_secondary = mirror_reflectivity
        # testeff does not support 2D distributions
        if self._telescope_model.is_file_2d("mirror_reflectivity"):
            mirror_reflectivity = self._get_one_dim_distribution(
                "mirror_reflectivity", "primary_mirror_incidence_angle"
            )
            mirror_reflectivity_secondary = self._get_one_dim_distribution(
                "mirror_reflectivity", "secondary_mirror_incidence_angle"
            )

        command = str(self._simtel_path.joinpath("sim_telarray/testeff"))
        if self.skip_correction_to_nsb_spectrum:
            command += " -nc"  # Do not apply correction to original altitude where B&E was derived
        command += " -I"  # Clear the fall-back configuration directories
        command += f" -I{self._telescope_model.config_file_directory}"
        if self.nsb_spectrum is not None:
            command += f" -fnsb {self.nsb_spectrum}"
        command += " -nm -nsb-extra"
        command += f" -alt {self._site_model.get_parameter_value('corsika_observation_level')}"
        command += f" -fatm {self._site_model.get_parameter_value('atmospheric_transmission')}"
        command += f" -flen {focal_length}"
        command += f" -fcur {curvature_radius:.3f}"
        command += f" {pixel_shape_cmd} {pixel_diameter}"
        if mirror_class == 0:
            command += f" -fmir {self._telescope_model.get_parameter_value('mirror_list')}"
        if mirror_class == 2:
            command += f" -fmir {self._telescope_model.get_parameter_value('fake_mirror_list')}"
        command += f" -fref {mirror_reflectivity}"
        if mirror_class == 2:
            command += " -m2"
            command += f" -fref2 {mirror_reflectivity_secondary}"
        command += " -teltrans "
        command += f"{self._telescope_model.get_parameter_value('telescope_transmission')[0]}"
        command += f" -camtrans {camera_transmission}"
        command += f" -fflt {camera_filter_file}"
        command += (
            f" -fang {self._telescope_model.camera.get_lightguide_efficiency_angle_file_name()}"
        )
        command += (
            f" -fwl {self._telescope_model.camera.get_lightguide_efficiency_wavelength_file_name()}"
        )
        command += f" -fqe {self._telescope_model.get_parameter_value('quantum_efficiency')}"
        command += " 200 1000"  # lmin and lmax
        command += " 300"  # Xmax
        command += f" {self._site_model.get_parameter_value('atmospheric_profile')}"
        command += f" {self.zenith_angle}"

        # Remove the default sim_telarray configuration directories
        command = general.clear_default_sim_telarray_cfg_directories(command)

        return (
            f"cd {self._simtel_path.joinpath('sim_telarray')} && {command}",
            self._file_simtel,
            self._file_log,
        )

    def _check_run_result(self, run_number=None):  # pylint: disable=unused-argument
        """Check run results.

        Raises
        ------
        RuntimeError
            if camera efficiency simulation results file does not exist.
        """
        # Checking run
        if not self._file_simtel.exists():
            msg = f"Camera efficiency simulation results file does not exist ({self._file_simtel})."
            self._logger.error(msg)
            raise RuntimeError(msg)

        self._logger.debug("Everything looks fine with output file.")

    def _get_one_dim_distribution(self, two_dim_parameter, weighting_distribution_parameter):
        """
        Calculate an average one-dimensional curve for testeff from the two-dimensional curve.

        The two-dimensional distribution is provided in two_dim_parameter. The distribution
        of weights to use for averaging the two-dimensional distribution is given in
        weighting_distribution_parameter.

        Parameters
        ----------
        two_dim_parameter: str
            The name of the two-dimensional distribution parameter.
        weighting_distribution_parameter: str
            The name of the parameter with the distribution of weights.

        Returns
        -------
        one_dim_file: Path
            The file path and name with the new one-dimensional distribution
        """
        incidence_angle_distribution_file = self._telescope_model.get_parameter_value(
            weighting_distribution_parameter
        )
        incidence_angle_distribution = self._telescope_model.read_incidence_angle_distribution(
            incidence_angle_distribution_file
        )
        self._logger.warning(
            f"The {' '.join(two_dim_parameter.split('_'))} distribution "
            "is a 2D one which testeff does not support. "
            "Instead of using the 2D distribution, the two dimensional distribution "
            "will be averaged, using the photon incidence angle distribution as weights. "
            "The incidence angle distribution is taken "
            f"from the file - {incidence_angle_distribution_file})."
        )
        two_dim_distribution = self._telescope_model.read_two_dim_wavelength_angle(
            self._telescope_model.get_parameter_value(two_dim_parameter)
        )
        distribution_to_export = self._telescope_model.calc_average_curve(
            two_dim_distribution, incidence_angle_distribution
        )
        new_file_name = (
            f"weighted_average_1D_{weighting_distribution_parameter}"
            f"_{self._telescope_model.get_parameter_value(two_dim_parameter)}"
        )
        return self._telescope_model.export_table_to_model_directory(
            new_file_name, distribution_to_export
        )

    def _validate_or_fix_nsb_spectrum_file_format(self, nsb_spectrum_file):
        """
        Validate or fix the nsb spectrum file format.

        The nsb spectrum file format required by sim_telarray has three columns:
        wavelength (nm), ignored, NSB flux [1e9 * ph/m2/s/sr/nm],
        where the second column is ignored by sim_telarray and the third is used for the NSB flux.
        This function makes sure the file has at least three columns,
        by copying the second column to the third.

        Parameters
        ----------
        nsb_spectrum_file: str or Path
            The path to the nsb spectrum file.

        Returns
        -------
        validated_nsb_spectrum_file: Path
            The path to the validated nsb spectrum file.
        """
        validated_nsb_spectrum_file = (
            self._telescope_model.config_file_directory / Path(nsb_spectrum_file).name
        )

        lines = ascii_handler.read_file_encoded_in_utf_or_latin(nsb_spectrum_file)

        with open(validated_nsb_spectrum_file, "w", encoding="utf-8") as file:
            for line in lines:
                if line.startswith("#"):
                    file.write(line)
                    continue
                split_line = line.split()
                if len(split_line) == 2:
                    split_line.append(split_line[1])
                    file.write(f"{split_line[0]} {split_line[1]} {split_line[2]}\n")
                else:
                    file.write(line)
        return validated_nsb_spectrum_file

    def _get_curvature_radius(self, mirror_class=1):
        """Get radius of curvature of dish."""
        if mirror_class == 2:
            return (
                self._telescope_model.get_parameter_value_with_unit("primary_mirror_diameter")
                .to("m")
                .value
            )

        if self._telescope_model.get_parameter_value("parabolic_dish"):
            return (
                2.0
                * self._telescope_model.get_parameter_value_with_unit("dish_shape_length")
                .to("m")
                .value
            )

        return (
            self._telescope_model.get_parameter_value_with_unit("dish_shape_length").to("m").value
        )



"""Command line parser for applications."""

import argparse
import logging
import re
from pathlib import Path

import astropy.units as u

import simtools.version
from simtools.utils import names


class CommandLineParser(argparse.ArgumentParser):
    """
    Command line parser for applications.

    Wrapper around standard python argparse.ArgumentParser.

    Command line arguments should be given in snake_case, e.g. 'input_meta'.

    Parameters
    ----------
    argparse.ArgumentParser class
        Object for parsing command line strings into Python objects. For a list of keywords, please\
        refer to argparse.ArgumentParser documentation.
    """

    def initialize_default_arguments(
        self,
        paths=True,
        output=False,
        simulation_model=None,
        simulation_configuration=None,
        db_config=False,
    ):
        """
        Initialize default arguments used by all applications (e.g., log level or test flag).

        Parameters
        ----------
        paths: bool
            Add path configuration to list of args.
        output: bool
            Add output file configuration to list of args.
        simulation_model: list
            List of simulation model configuration parameters to add to list of args
            (use: 'model_version', 'telescope', 'site')
        simulation_configuration: dict
            Dict of simulation software configuration parameters to add to list of args.
        db_config: bool
            Add database configuration parameters to list of args.
        """
        self.initialize_simulation_model_arguments(simulation_model)
        self.initialize_simulation_configuration_arguments(simulation_configuration)
        if db_config:
            self.initialize_db_config_arguments()
        if paths:
            self.initialize_path_arguments()
        if output:
            self.initialize_output_arguments()
        self.initialize_config_files()
        self.initialize_application_execution_arguments()
        self.initialize_user_arguments()

    def initialize_config_files(self):
        """Initialize configuration files."""
        _job_group = self.add_argument_group("configuration")
        _job_group.add_argument(
            "--config",
            help="simtools configuration file",
            default=None,
            type=str,
            required=False,
        )
        _job_group.add_argument(
            "--env_file",
            help="file with environment variables",
            default=".env",
            type=str,
            required=False,
        )

    def initialize_path_arguments(self):
        """Initialize paths."""
        _job_group = self.add_argument_group("paths")
        _job_group.add_argument(
            "--data_path",
            help="path pointing towards data directory",
            type=Path,
            default="./data/",
            required=False,
        )
        _job_group.add_argument(
            "--output_path",
            help="path pointing towards output directory",
            type=Path,
            default="./simtools-output/",
            required=False,
        )
        _job_group.add_argument(
            "--model_path",
            help="path pointing towards simulation model file directory",
            type=Path,
            default="./",
            required=False,
        )
        _job_group.add_argument(
            "--simtel_path",
            help="path pointing to sim_telarray installation",
            type=Path,
            required=False,
        )

    def initialize_output_arguments(self):
        """Initialize application output files(s)."""
        _job_group = self.add_argument_group("output")
        _job_group.add_argument(
            "--output_file",
            help="output data file",
            type=str,
            required=False,
        )
        _job_group.add_argument(
            "--output_file_format",
            help="file format of output data",
            type=str,
            default="ecsv",
            required=False,
        )
        _job_group.add_argument(
            "--skip_output_validation",
            help="skip output data validation against schema",
            default=False,
            required=False,
            action="store_true",
        )

    def initialize_application_execution_arguments(self):
        """Initialize application execution arguments."""
        _job_group = self.add_argument_group("execution")
        _job_group.add_argument(
            "--test",
            help="test option for faster execution during development",
            action="store_true",
            required=False,
        )
        _job_group.add_argument(
            "--label",
            help="job label",
            required=False,
        )
        _job_group.add_argument(
            "--log_level",
            action="store",
            default="info",
            help="log level to print",
            required=False,
        )
        _job_group.add_argument(
            "--version", action="version", version=f"%(prog)s {simtools.version.__version__}"
        )

    def initialize_user_arguments(self):
        """Initialize user arguments."""
        _job_group = self.add_argument_group("user")
        _job_group.add_argument(
            "--user_name",
            help="user name",
            type=str,
            required=False,
        )
        _job_group.add_argument(
            "--user_organization",
            help="user organization",
            type=str,
            required=False,
        )
        _job_group.add_argument(
            "--user_email",
            help="user email",
            type=str,
            required=False,
        )
        _job_group.add_argument(
            "--user_orcid",
            help="user ORCID",
            type=str,
            required=False,
        )

    def initialize_db_config_arguments(self):
        """Initialize DB configuration parameters."""
        _job_group = self.add_argument_group("database configuration")
        _job_group.add_argument("--db_api_user", help="database user", type=str, required=False)
        _job_group.add_argument("--db_api_pw", help="database password", type=str, required=False)
        _job_group.add_argument("--db_api_port", help="database port", type=int, required=False)
        _job_group.add_argument(
            "--db_server", help="database server address", type=str, required=False
        )
        _job_group.add_argument(
            "--db_api_authentication_database",
            help="database  with user info (optional)",
            type=str,
            required=False,
            default=None,
        )
        _job_group.add_argument(
            "--db_simulation_model",
            help="name of simulation model database",
            type=str.strip,
            required=False,
            default=None,
        )
        _job_group.add_argument(
            "--db_simulation_model_version",
            help="version of simulation model database",
            type=str.strip,
            required=False,
            default=None,
        )

    def initialize_simulation_model_arguments(self, model_options):
        """
        Initialize default arguments for simulation model definition.

        Note that the model version is always required.

        Parameters
        ----------
        model_options: list
            Options to be set: "telescope", "site", "layout", "layout_file"
        """
        if model_options is None:
            return

        _job_group = self.add_argument_group("simulation model")
        if "model_version" in model_options:
            _job_group.add_argument(
                "--model_version",
                help="production model version",
                type=str,
                default=None,
                nargs="+",
            )
        if "parameter_version" in model_options:
            _job_group.add_argument(
                "--parameter_version",
                help="model parameter version",
                type=str,
                default=None,
            )
        _job_group.add_argument(
            "--overwrite_model_parameters",
            help="File name to overwrite model parameters from DB with provided values",
            type=str,
            required=False,
        )

        if any(
            option in model_options for option in ["site", "telescope", "layout", "layout_file"]
        ):
            self._add_model_option_site(_job_group)

        if "telescope" in model_options:
            _job_group.add_argument(
                "--telescope",
                help="telescope model name (e.g., LSTN-01, SSTS-design, ...)",
                type=self.telescope,
            )
        if "layout" in model_options or "layout_file" in model_options:
            _job_group = self._add_model_option_layout(
                job_group=_job_group,
                model_options=model_options,
                # layout info is always required for layout related tasks with the exception
                # of listing the available layouts in the DB
                required="--list_available_layouts" not in self._option_string_actions,
            )

    def initialize_simulation_configuration_arguments(self, simulation_configuration):
        """
        Initialize default arguments for simulation configuration and simulation software.

        Parameters
        ----------
        simulation_configuration: dict
            Dict of simulation software configuration parameters.
        """
        if simulation_configuration is None:
            return

        if "software" in simulation_configuration:
            self._initialize_simulation_software()
        if "corsika_configuration" in simulation_configuration:
            self._initialize_simulation_configuration(
                group_name="simulation configuration",
                selected_parameters=simulation_configuration["corsika_configuration"],
                available_parameters=self._get_dictionary_with_corsika_configuration(),
            )
            self._initialize_simulation_configuration(
                group_name="shower parameters",
                selected_parameters=simulation_configuration["corsika_configuration"],
                available_parameters=self._get_dictionary_with_shower_configuration(),
            )
        if "sim_telarray_configuration" in simulation_configuration:
            self._initialize_simulation_configuration(
                group_name="sim_telarray configuration",
                selected_parameters=simulation_configuration["sim_telarray_configuration"],
                available_parameters=self._get_dictionary_with_sim_telarray_configuration(),
            )

    def _initialize_simulation_software(self):
        """Initialize simulation software arguments."""
        _software_group = self.add_argument_group("simulation software")
        _software_group.add_argument(
            "--simulation_software",
            help="Simulation software steps.",
            type=str,
            choices=["corsika", "sim_telarray", "corsika_sim_telarray"],
            required=True,
            default="corsika_sim_telarray",
        )

    @staticmethod
    def _get_dictionary_with_corsika_configuration():
        """Return dictionary with CORSIKA configuration parameters."""
        from simtools.corsika.primary_particle import PrimaryParticle  # pylint: disable=C0415

        return {
            "primary": {
                "help": (
                    "Primary particle to simulate. "
                    "(choices for common names: "
                    f"{', '.join(PrimaryParticle.particle_names().keys())}; "
                    "use '--primary_ID_type' to use other particle ID types)."
                ),
                "type": str.lower,
                "required": True,
            },
            "primary_id_type": {
                "help": "Primary particle ID type",
                "type": str,
                "required": False,
                "choices": ["common_name", "corsika7_id", "pdg_id"],
                "default": "common_name",
            },
            "azimuth_angle": {
                "help": (
                    "Telescope pointing direction in azimuth. "
                    "It can be in degrees between 0 and 360 or one of north, south, east or west. "
                    "North is 0 degrees and the azimuth grows clockwise (East is 90 degrees)."
                ),
                "type": CommandLineParser.azimuth_angle,
                "required": True,
                "default": 0 * u.deg,
            },
            "zenith_angle": {
                "help": "Zenith angle in degrees (between 0 and 180).",
                "type": CommandLineParser.zenith_angle,
                "required": True,
                "default": 20 * u.deg,
            },
            "nshow": {
                "help": "Number of showers per run to simulate.",
                "type": int,
                "required": False,
            },
            "run_number_offset": {
                "help": "An offset for the run number to be simulated.",
                "type": int,
                "required": False,
                "default": 0,
            },
            "run_number": {
                "help": "Run number to be simulated.",
                "type": int,
                "required": True,
                "default": 1,
            },
            "event_number_first_shower": {
                "help": "Event number of first shower",
                "type": int,
                "required": False,
                "default": 1,
            },
            "correct_for_b_field_alignment": {
                "help": "Correct for B-field alignment",
                "action": "store_true",
                "required": False,
                "default": True,
            },
            "curved_atmosphere_min_zenith_angle": {
                "help": (
                    "Minimum zenith angle (deg) for using curved-atmosphere CORSIKA binaries. "
                ),
                "type": CommandLineParser.zenith_angle,
                "required": False,
                "default": 65 * u.deg,
            },
        }

    @staticmethod
    def _get_dictionary_with_shower_configuration():
        """Return dictionary with shower configuration parameters."""
        return {
            "eslope": {
                "help": "Slope of the energy spectrum.",
                "type": float,
                "required": False,
                "default": -2.0,
            },
            "energy_range": {
                "help": (
                    "Energy range of the primary particle (min/max value, e'g', '10 GeV 5 TeV')."
                ),
                "type": CommandLineParser.parse_quantity_pair,
                "required": False,
                "default": ["3 GeV 330 TeV"],
            },
            "view_cone": {
                "help": (
                    "View cone radius for primary arrival directions "
                    "(min/max value, e.g. '0 deg 10 deg')."
                ),
                "type": CommandLineParser.parse_quantity_pair,
                "required": False,
                "default": ["0 deg 0 deg"],
            },
            "core_scatter": {
                "help": "Scatter radius for shower cores (number of use; scatter radius).",
                "type": CommandLineParser.parse_integer_and_quantity,
                "required": False,
                "default": ["10 1400 m"],
            },
        }

    @staticmethod
    def _get_dictionary_with_sim_telarray_configuration():
        """Return dictionary with sim_telarray configuration parameters."""
        return {
            "sim_telarray_instrument_seeds": {
                "help": (
                    "Random seed used for sim_telarray instrument setup. "
                    "If '--sim_telarray_random_instrument_instances' is not set: "
                    "use as sim_telarray seed ('random_seed' parameter). Otherwise: "
                    "use as base seed to generate the random instrument instance seeds."
                ),
                "type": str,
                "required": False,
            },
            "sim_telarray_random_instrument_instances": {
                "help": "Number of random instrument instances initialized in sim_telarray.",
                "type": int,
                "required": False,
            },
        }

    def _initialize_simulation_configuration(
        self, group_name, selected_parameters, available_parameters
    ):
        """
        Initialize simulation configuration arguments.

        Parameters
        ----------
        group_name : str
            Name of the group of arguments.
        selected_parameters : list
            List of selected parameters to be added to the group.
        available_parameters : dict
            Dictionary with available parameters and their configuration.
        """
        configuration_group = self.add_argument_group(group_name)

        if "all" in selected_parameters:
            selected_parameters = available_parameters.keys()

        for param in selected_parameters:
            try:
                configuration_group.add_argument(f"--{param}", **available_parameters[param])
            except KeyError:
                pass

    @staticmethod
    def _add_model_option_layout(job_group, model_options, required=True):
        """
        Add layout option to the job group.

        Parameters
        ----------
        job_group: argparse.ArgumentParser
            Job group
        model_options: list
            List of model options.

        Returns
        -------
        argparse.ArgumentParser
        """
        _layout_group = job_group.add_mutually_exclusive_group(required=required)
        _layout_group.add_argument(
            "--array_layout_name",
            help="array layout name(s) (e.g., alpha, subsystem_msts)",
            nargs="+",
            type=str,
            required=False,
        )
        _layout_group.add_argument(
            "--array_element_list",
            help="list of array elements (e.g., LSTN-01, LSTN-02, MSTN).",
            nargs="+",
            type=str,
            required=False,
            default=None,
        )
        if "layout_file" in model_options:
            _layout_group.add_argument(
                "--array_layout_file",
                help="file(s) with the list of array elements (astropy table format).",
                nargs="+",
                type=str,
                required=False,
                default=None,
            )
        if "layout_parameter_file" in model_options:
            _layout_group.add_argument(
                "--array_layout_parameter_file",
                help="Array layout model parameter file (typically in JSON format).",
                type=str,
                required=False,
                default=None,
            )
        if "plot_all_layouts" in model_options:
            _layout_group.add_argument(
                "--plot_all_layouts",
                help="plot all available layouts",
                action="store_true",
                required=False,
            )
        return job_group

    def _add_model_option_site(self, job_group):
        """
        Add site option to the job group.

        Parameters
        ----------
        job_group: argparse.ArgumentParser
            Job group

        Returns
        -------
        argparse.ArgumentParser
        """
        job_group.add_argument(
            "--site", help="site (e.g., North, South)", type=self.site, required=False
        )
        return job_group

    @staticmethod
    def scientific_int(value):
        """
        Convert string (including scientific notation) to integer.

        Parameters
        ----------
        value: str or int or float
            Value to convert to integer. Can be a regular integer,
            float, or string in scientific notation (e.g., '1e7').

        Returns
        -------
        int
            Converted integer value

        Raises
        ------
        argparse.ArgumentTypeError
            If the value cannot be converted to an integer
        """
        try:
            f = float(value)
            if not f.is_integer():
                raise ValueError
            return int(f)
        except (ValueError, TypeError) as exc:
            raise argparse.ArgumentTypeError(
                f"Invalid integer value: '{value}'. "
                "Expected an integer or scientific notation like '1e7'."
            ) from exc

    @staticmethod
    def site(value):
        """
        Argument parser type to check that a valid site name is given.

        Parameters
        ----------
        value: str
            site name

        Returns
        -------
        str
            Validated site name

        Raises
        ------
        argparse.ArgumentTypeError
            for invalid sites
        """
        names.validate_site_name(str(value))
        return str(value)

    @staticmethod
    def telescope(value):
        """
        Argument parser type to check that a valid telescope name is given.

        Parameters
        ----------
        value: str
            telescope name

        Returns
        -------
        str
            Validated telescope name

        Raises
        ------
        argparse.ArgumentTypeError
            for invalid telescope
        """
        names.validate_array_element_name(str(value))
        return str(value)

    @staticmethod
    def efficiency_interval(value):
        """
        Argument parser type to check that value is an efficiency in the interval [0,1].

        Parameters
        ----------
        value: float
            value provided through the command line

        Returns
        -------
        float
            Validated efficiency interval

        Raises
        ------
        argparse.ArgumentTypeError
            When value is outside of the interval [0,1]
        """
        fvalue = float(value)
        if fvalue < 0.0 or fvalue > 1.0:
            raise argparse.ArgumentTypeError(f"{value} outside of allowed [0,1] interval")

        return fvalue

    @staticmethod
    def zenith_angle(angle):
        """
        Argument parser type to check that the zenith angle provided is in the interval [0, 180].

        We allow here zenith angles larger than 90 degrees in the improbable case
        such simulations are requested. It is not guaranteed that the actual simulation software
        supports such angles!.

        Parameters
        ----------
        angle: float, str, astropy.Quantity
            zenith angle to verify

        Returns
        -------
        Astropy.Quantity
            Validated zenith angle in degrees

        Raises
        ------
        argparse.ArgumentTypeError
            When angle is outside of the interval [0, 180]
        """
        logger = logging.getLogger(__name__)

        try:
            try:
                fangle = float(angle) * u.deg
            except ValueError:
                fangle = u.Quantity(angle).to("deg")
        except TypeError as exc:
            logger.error(
                "The zenith angle provided is not a valid numerical or astropy.Quantity value."
            )
            raise exc
        if fangle < 0.0 * u.deg or fangle > 180.0 * u.deg:
            raise argparse.ArgumentTypeError(
                f"The provided zenith angle, {angle:.1f}, "
                "is outside of the allowed [0, 180] interval"
            )
        return fangle

    @staticmethod
    def azimuth_angle(angle):
        """
        Argument parser type to check that the azimuth angle provided is in the interval [0, 360].

        Other allowed options are north, south, east or west which will be translated to an angle
        where north corresponds to zero.

        Parameters
        ----------
        angle: float or str
            azimuth angle to verify or convert

        Returns
        -------
        Astropy.Quantity
            Validated/Converted azimuth angle in degrees

        Raises
        ------
        argparse.ArgumentTypeError
            When angle is outside of the interval [0, 360] or not in (north, south, east, west)
        """
        logger = logging.getLogger(__name__)
        try:
            fangle = float(angle)
            if fangle < 0.0 or fangle > 360.0:
                raise argparse.ArgumentTypeError(
                    f"The provided azimuth angle, {angle:.1f}, "
                    "is outside of the allowed [0, 360] interval"
                )

            return fangle * u.deg
        except ValueError:
            logger.debug(
                "The azimuth angle provided is not a valid numeric value. "
                "Will check if it is an astropy.Quantity instead"
            )
        except TypeError as exc:
            logger.error("The azimuth angle provided is not a valid numerical or string value.")
            raise exc
        try:
            return u.Quantity(angle).to("deg")
        except TypeError:
            logger.debug(
                "The azimuth angle provided is not a valid astropy.Quantity. "
                "Will check if it is (north, south, east, west) instead"
            )
        azimuth_map = {
            "north": 0 * u.deg,
            "south": 180 * u.deg,
            "east": 90 * u.deg,
            "west": 270 * u.deg,
        }
        azimuth_angle = angle.lower()
        if azimuth_angle in azimuth_map:
            return azimuth_map[azimuth_angle]
        raise argparse.ArgumentTypeError(
            "The azimuth angle given as string can only be one of "
            f"(north, south, east, west), not {angle}. Otherwise use numerical values."
        )

    @staticmethod
    def parse_quantity_pair(string):
        """
        Parse a string representing a pair of astropy quantities.

        Returns
        -------
        tuple
            A tuple of two astropy.units.Quantity objects.

        Raises
        ------
        ValueError
            If the string cannot be parsed into exactly two quantities.
        """
        pattern = r"(?>[\d\.eE+-]+)\s*(?>[A-Za-z]+)"
        matches = re.findall(pattern, string)
        if len(matches) != 2:
            raise ValueError("Input string does not contain exactly two quantities.")

        try:
            return tuple(u.Quantity(m) for m in matches)
        except Exception as exc:
            raise ValueError(f"Could not parse quantities: {exc}") from exc

    @staticmethod
    def parse_integer_and_quantity(input_string):
        """
        Parse a string representing an integer and a quantity with units.

        This is e.g., used for the 'core_scatter' argument.

        Parameters
        ----------
        input_string: str
            The input string (e.g., "5 1500 m") or
            a tuple converted to string (e.g., "(5, <Quantity 1500 m>)").

        Returns
        -------
        tuple: A tuple containing an integer and an astropy.units.Quantity object.

        Raises
        ------
        ValueError: If the input string does not match the required format.
        """
        # tuple converted to string: "(5, <Quantity 1500 m>)"
        if all(char in input_string for char in ["(", ")", ","]):
            pattern = r"\((\d+), <Quantity ([\d.]+) (.+)>\)"
            match = re.match(pattern, input_string)
        # string with integer and quantity: "5 1500 m"
        else:
            pattern = r"(\d+)\s+(\d+\.?\d*)\s*([a-zA-Z]+)"
            match = re.match(pattern, input_string.strip())

        if not match:
            raise ValueError("Input string does not contain an integer and a astropy quantity.")

        return (int(match.group(1)), u.Quantity(float(match.group(2)), match.group(3)))



"""Upload a simulation model (parameters and production tables) to the database."""

import logging
import shutil
from pathlib import Path

from packaging.version import Version

from simtools.io import ascii_handler
from simtools.job_execution.job_manager import retry_command
from simtools.utils import names

logger = logging.getLogger(__name__)


def add_complete_model(
    tmp_dir,
    db,
    db_simulation_model,
    db_simulation_model_version,
    repository_url,
    repository_branch=None,
):
    """
    Upload a complete model including model parameters and production tables to the database.

    Generate compound indexes for the new database.

    Parameters
    ----------
    tmp_dir : Path or str
        Temporary directory to use for cloning the repository.
    db : DatabaseHandler
        Database handler object.
    db_simulation_model : str
        Name of the simulation model in the database.
    db_simulation_model_version : str
        Version of the simulation model in the database.
    repository_url : str
        URL of the simulation model repository to clone.
    repository_branch : str, optional
        Branch of the repository to use. If None, the default branch is used.

    Returns
    -------
    None
        This function does not return a value.
    """
    if not _confirm_remote_database_upload(db):
        return

    repository_dir = None
    try:
        repository_dir = clone_simulation_model_repository(
            tmp_dir,
            repository_url,
            db_simulation_model_version=db_simulation_model_version,
            repository_branch=repository_branch,
        )

        add_model_parameters_to_db(
            input_path=repository_dir / "simulation-models" / "model_parameters", db=db
        )

        add_production_tables_to_db(
            input_path=repository_dir / "simulation-models" / "productions", db=db
        )

        db.generate_compound_indexes_for_databases(
            db_name=None,
            db_simulation_model=db_simulation_model,
            db_simulation_model_version=db_simulation_model_version,
        )
    except Exception as exc:
        raise RuntimeError(f"Upload of simulation model failed: {exc}") from exc
    finally:
        if repository_dir is not None and repository_dir.exists():
            shutil.rmtree(repository_dir)

    logger.info("Upload of simulation model completed successfully")


def add_values_from_json_to_db(file, collection, db, file_prefix):
    """
    Upload new model parameter from json files to db.

    Parameters
    ----------
    file : list
        JSON file to be uploaded to the DB.
    collection : str
        The DB collection to which to add the file.
    db : DatabaseHandler
        Database handler object.
    file_prefix : str
        Path to location of all additional files to be uploaded.
    """
    par_dict = ascii_handler.collect_data_from_file(file_name=file)
    logger.debug(
        f"Adding the following parameter to the DB: {par_dict['parameter']} "
        f"version {par_dict['parameter_version']} "
        f"(collection {collection} in database {db.get_db_name()})"
    )

    db.add_new_parameter(
        par_dict=par_dict,
        collection_name=collection,
        file_prefix=file_prefix,
    )


def add_model_parameters_to_db(input_path, db):
    """
    Read model parameters from a directory and upload them to the database.

    Parameters
    ----------
    input_path : Path, str
        Path to the directory containing the model parameters.
    db : DatabaseHandler
        Database handler object.
    """
    input_path = Path(input_path)
    logger.info(f"Reading model parameters from repository path {input_path}")
    for element in filter(Path.is_dir, input_path.iterdir()):
        collection = names.get_collection_name_from_array_element_name(element.name, False)
        if collection == "Files":
            logger.info("Files (tables) are uploaded with the corresponding model parameters")
            continue
        logger.info(f"Reading model parameters for {element.name} into collection {collection}")
        files_to_insert = list(Path(element).rglob("*json"))
        for file in files_to_insert:
            add_values_from_json_to_db(
                file=file,
                collection=collection,
                db=db,
                file_prefix=input_path / "Files",
            )


def add_production_tables_to_db(input_path, db):
    """
    Read production tables from a directory and upload them to the database.

    One dictionary per collection is prepared for each model version, containing
    tables of all array elements, sites, and configuration parameters.

    Parameters
    ----------
    input_path : Path, str
        Path to the directory containing the production tables.
    db : DatabaseHandler
        Database handler object.
    """
    input_path = Path(input_path)
    logger.info(f"Reading production tables from repository path {input_path}")

    for model in sorted(filter(Path.is_dir, input_path.iterdir())):
        logger.info(f"Reading production tables for model version {model.name}")
        model_dict = _read_production_tables(model)

        for collection, data in model_dict.items():
            if data["parameters"]:
                logger.info(
                    f"Adding production table for {collection} "
                    f"(model version {model.name}) to the database"
                )
                db.add_production_table(production_table=data)
            else:
                logger.info(f"No production table for {collection} in model version {model.name}")


def _read_production_tables(model_path):
    """
    Read production tables from a directory.

    Take into account that some productions include patch updates only. Read in this cases
    all models from the model version history, starting with the earliest one.

    Parameters
    ----------
    model_path : Path
        Path to the directory containing the production tables for a specific model version.
    """
    model_dict = {}
    models = [model_path.name]
    if (model_path / "info.yml").exists():
        info = ascii_handler.collect_data_from_file(file_name=model_path / "info.yml")
        if info.get("model_update") == "patch_update":
            models.extend(info.get("model_version_history", []))
    # sort oldest --> newest
    models = sorted(set(models), key=Version, reverse=False)
    for model in models:
        for file in sorted((model_path.parent / model).rglob("*json")):
            _read_production_table(model_dict, file, model)

    # ensure that the for patch updates the model version is set correctly
    for table in model_dict.values():
        table["model_version"] = model_path.name

    _remove_deprecated_model_parameters(model_dict)

    return model_dict


def _read_production_table(model_dict, file, model_name):
    """Read a single production table from file."""
    array_element = file.stem
    collection = names.get_collection_name_from_array_element_name(array_element, False)
    model_dict.setdefault(
        collection,
        {
            "collection": collection,
            "model_version": model_name,
            "parameters": {},
            "design_model": {},
            "deprecated_parameters": [],
        },
    )
    parameter_dict = ascii_handler.collect_data_from_file(file_name=file)
    logger.debug(
        f"Reading production table for {array_element} "
        f"(model_version {model_name}, collection {collection})"
    )
    try:
        if array_element in ("configuration_corsika", "configuration_sim_telarray"):
            model_dict[collection]["parameters"] = parameter_dict["parameters"]
        else:
            model_dict[collection]["parameters"].setdefault(array_element, {}).update(
                parameter_dict["parameters"][array_element]
            )
    except KeyError as exc:
        logger.error(f"KeyError: {exc}")
        raise

    try:
        model_dict[collection]["design_model"][array_element] = parameter_dict["design_model"][
            array_element
        ]
    except KeyError:
        pass

    try:
        model_dict[collection]["deprecated_parameters"] = parameter_dict["deprecated_parameters"]
    except KeyError:
        pass

    model_dict[collection]["model_version"] = model_name


def _remove_deprecated_model_parameters(model_dict):
    """
    Remove deprecated parameters from all tables in a model dictionary.

    Parameters
    ----------
    model_dict : dict
        Production tables for a specific model version.
    """
    for table in model_dict.values():
        for params in table.get("parameters", {}).values():
            for param in table.get("deprecated_parameters", []):
                if param in params:
                    logger.info(
                        f"Deprecated parameter {param} in production table {table['collection']}"
                    )
                    params.pop(param)


def _confirm_remote_database_upload(db):
    """
    Confirm with the user that they want to upload to a remote database.

    Returns True if the user confirms, False otherwise. Returns also True
    if the database is not remote.

    Parameters
    ----------
    db : DatabaseHandler
        Database handler object.

    Returns
    -------
    bool
        True if user confirms upload, False otherwise.
    """
    abort_message = "Operation aborted."

    if not db.is_remote_database():
        return True

    db_config = db.db_config
    db_server = db_config.get("db_server", "unknown server") if db_config else "unknown server"

    try:
        # First confirmation
        user_input = input(
            f"Do you really want to upload to remote DB {db_server}? Type 'yes' to confirm: "
        )
        if user_input != "yes":
            logger.info(abort_message)
            return False

        # Second confirmation
        user_input = input(
            f"Let be sure: do you really want to upload to remote DB {db_server}? "
            "Type 'yes' to confirm: "
        )
        if user_input != "yes":
            logger.info(abort_message)
            return False

        return True

    except (EOFError, KeyboardInterrupt):
        logger.info(abort_message)
        return False


def clone_simulation_model_repository(
    target_dir, repository_url, db_simulation_model_version, repository_branch
):
    """
    Clone a git repository containing simulation model parameters and production tables.

    Parameters
    ----------
    target_dir : Path or str
        Target directory to clone the repository into.
    repository_url : str
        URL of the git repository to clone.
    db_simulation_model_version : str
        Version tag of the simulation model to checkout.
    repository_branch : str, optional
        Branch of the repository to use. If None, the version tag is used.

    Returns
    -------
    Path
        Path to the cloned repository.
    """
    if repository_branch:
        logger.info(f"Using branch: {repository_branch}")
    else:
        logger.info(f"Using version tag: {db_simulation_model_version}")

    target_dir = Path(target_dir)
    target_dir = target_dir if target_dir.is_absolute() else Path.cwd() / target_dir

    if target_dir.exists():
        shutil.rmtree(target_dir)

    logger.info(f"Cloning model parameters from {repository_url}")

    if repository_branch:
        command = f'git clone --depth=1 -b "{repository_branch}" "{repository_url}" "{target_dir}"'
    else:
        # Use version tag with detached head (generates warning but that's fine)
        command = (
            f'git clone --branch "{db_simulation_model_version}" '
            f'--depth 1 "{repository_url}" "{target_dir}"'
        )

    retry_command(command)

    return target_dir



#!/usr/bin/python3

r"""
Produces a markdown file for a given array element, site, and model version.

The report includes detailed information on each parameter,
such as the parameter name, value, unit, description, and short description.
"""

from pathlib import Path

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.reporting.docs_auto_report_generator import ReportGenerator
from simtools.reporting.docs_read_parameters import ReadParameters


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=("Produce a markdown report for model parameters."),
    )

    config.parser.add_argument(
        "--all_telescopes",
        action="store_true",
        help="Produce reports for all telescopes.",
    )

    config.parser.add_argument(
        "--all_model_versions",
        action="store_true",
        help="Produce reports for all model versions.",
    )

    config.parser.add_argument(
        "--all_sites", action="store_true", help="Produce reports for all sites."
    )

    config.parser.add_argument(
        "--observatory",
        action="store_true",
        help="Produce reports for an observatory at a given site.",
    )

    return config.initialize(
        db_config=True, simulation_model=["site", "telescope", "model_version"]
    )


def main():
    """Produce a markdown file for a given array element, site, and model version."""
    app_context = startup_application(_parse)
    output_path = app_context.io_handler.get_output_directory()

    if any(
        [
            app_context.args.get("all_telescopes"),
            app_context.args.get("all_sites"),
            app_context.args.get("all_model_versions"),
        ]
    ):
        ReportGenerator(
            app_context.db_config,
            app_context.args,
            output_path,
        ).auto_generate_array_element_reports()

    else:
        model_version = app_context.args["model_version"]
        ReadParameters(
            app_context.db_config,
            app_context.args,
            Path(output_path / f"{model_version}"),
        ).produce_array_element_report()

        app_context.logger.info(
            f"Markdown report generated for {app_context.args['site']}"
            f" Telescope {app_context.args['telescope']} (v{model_version}):"
            f" {output_path}"
        )


if __name__ == "__main__":
    main()
